'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import constants
from pm4py.util import vis_utils, exec_utils
import shutil
from pm4py.visualization.variants_duration.variants import classic
from enum import Enum
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
import pandas as pd


class Variants(Enum):
    CLASSIC = classic


def apply(log: Union[EventLog, pd.DataFrame], variant=Variants.CLASSIC, parameters: Optional[Dict[Any, Any]] = None):
    """
    Comparative visualization of process variants based on the average time between the different steps

    Parameters
    -----------------
    log
        Event log
    variant
        Variant of the algorithm to be used (Variants.CLASSIC)
    parameters
        Variant-specific parameters, including:
        - Parameters.ALIGNMENT_CRITERIA => specifies if variants should be aligned on the "start", "end", or a specific activity.
        - Parameters.MAX_VARIANTS => max variants to visualize in the graph
        - Parameters.FORMAT => the format of the visualization
        - Parameters.NODE_HEIGHT, Parameters.NODE_WIDTH, Parameters.EDGE_PENWIDTH, Parameters.MIN_HORIZONTAL_DISTANCE,
            Parameters.MAX_HORIZONTAL_DISTANCE, Parameters.LAYOUT_EXT_MULTIPLIER, Parameters.SHOW_LEGEND,
            Parameters.ENABLE_GRAPH_TITLE, Parameters.GRAPH_TITLE

    """
    if parameters is None:
        parameters = {}

    import pm4py
    var_paths_durs = pm4py.get_variants_paths_duration(log)

    return exec_utils.get_variant(variant).apply(var_paths_durs, parameters)


def view(figure: str):
    """
    Views the dotted chart on the screen

    Parameters
    ---------------
    figure
        Path to the dotted chart
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        if constants.DEFAULT_GVIZ_VIEW == "matplotlib_view":
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg

            img = mpimg.imread(figure)
            plt.axis("off")
            plt.tight_layout(pad=0, w_pad=0, h_pad=0)
            plt.imshow(img)
            plt.show()
            return

        if vis_utils.check_visualization_inside_jupyter():
            vis_utils.view_image_in_jupyter(figure)
        else:
            vis_utils.open_opsystem_image_viewer(figure)


def save(figure: str, output_file_path: str):
    """
    Saves the dotted chart to a specified path

    Parameters
    ----------------
    figure
        Current path to the dotted chart
    output_file_path
        Destination path
    """
    shutil.copyfile(figure, output_file_path)
    return ""


def serialize(figure: str):
    """
    Performs the serialization of the dotted chart visualization

    Parameters
    -----------------
    figure
        Current path to the dotted chart
    """
    with open(figure, "rb") as f:
        return f.read()


def matplotlib_view(figure: str):
    """
    Views the dotted chart on the screen using Matplotlib

    Parameters
    ---------------
    figure
        Path to the dotted chart
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg

        img = mpimg.imread(figure)
        plt.imshow(img)
        plt.show()



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any, Tuple, Set, List
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel import constants as ocel_constants
import pandas as pd
from pm4py.objects.ocel.obj import OCEL
from copy import copy


class Parameters(Enum):
    EVENT_ID = ocel_constants.PARAM_EVENT_ID
    OBJECT_ID = ocel_constants.PARAM_OBJECT_ID
    EVENT_ACTIVITY = ocel_constants.PARAM_EVENT_ACTIVITY
    PREFILTERING = "prefiltering"


def aggregate_events(
    associations: Dict[str, Set[Tuple[str, str]]]
) -> Dict[str, Set[str]]:
    """
    Utility method to calculate the "events" metric from the associations.
    """
    ret = {}
    for act in associations:
        ret[act] = set()
        for el in associations[act]:
            ret[act].add(el[0])
    return ret


def aggregate_unique_objects(
    associations: Dict[str, Set[Tuple[str, str]]]
) -> Dict[str, Set[str]]:
    """
    Utility method to calculate the "unique objects" metric from the associations.
    """
    ret = {}
    for act in associations:
        ret[act] = set()
        for el in associations[act]:
            ret[act].add(el[1])
    return ret


def aggregate_total_objects(
    associations: Dict[str, Set[Tuple[str, str]]]
) -> Dict[str, Set[Tuple[str, str]]]:
    """
    Utility method to calculate the "total objects" metric from the associations.
    """
    return associations


def find_associations_from_relations_df(
    relations_df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, List[Tuple[str, str]]]:
    """
    Associates each activity in the relationship dataframe with the combinations
    of event identifiers and objects that are associated to the activity.

    Parameters
    ------------------
    rel_df
        Relations dataframe
    parameters
        Parameters of the method, including:
        - Parameters.EVENT_ID => the attribute to use as event identifier
        - Parameters.OBJECT_ID => the attribute to use as object identifier
        - Parameters.EVENT_ACTIVITY => the attribute to use as activity

    Returns
    -----------------
    dict_associations
        Dictionary that associates each activity to its (ev. id, obj id.) combinations.
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel_constants.DEFAULT_EVENT_ID
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel_constants.DEFAULT_OBJECT_ID
    )
    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY,
        parameters,
        ocel_constants.DEFAULT_EVENT_ACTIVITY,
    )
    prefiltering = exec_utils.get_param_value(
        Parameters.PREFILTERING, parameters, "none"
    )

    if prefiltering == "start":
        relations_df = relations_df.groupby(object_id).first().reset_index()
    elif prefiltering == "end":
        relations_df = relations_df.groupby(object_id).last().reset_index()
    associations1 = (
        relations_df.groupby(event_activity)[[event_id, object_id]]
        .agg(list)
        .to_dict()
    )

    associations = {}
    for act, evs in associations1[event_id].items():
        associations[act] = []
        objs = associations1[object_id][act]
        for i in range(len(evs)):
            associations[act].append((evs[i], objs[i]))

    return associations


def find_associations_from_ocel(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, Set[Any]]:
    """
    Associates each activity in the OCEL with the combinations
    of event identifiers and objects that are associated to the activity.

    Parameters
    ------------------
    ocel
        Object-centric event log
    parameters
        Parameters of the method, including:
        - Parameters.EVENT_ID => the attribute to use as event identifier
        - Parameters.OBJECT_ID => the attribute to use as object identifier
        - Parameters.EVENT_ACTIVITY => the attribute to use as activity

    Returns
    -----------------
    dict_associations
        Dictionary that associates each activity to its (ev. id, obj id.) combinations.
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, ocel.event_activity
    )

    new_parameters = copy(parameters)
    new_parameters[Parameters.EVENT_ID] = event_id
    new_parameters[Parameters.OBJECT_ID] = object_id
    new_parameters[Parameters.EVENT_ACTIVITY] = event_activity

    return find_associations_from_relations_df(
        ocel.relations, parameters=new_parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.algo.conformance.tokenreplay.variants import token_replay
from pm4py.algo.evaluation.generalization.variants import (
    token_based as generalization_token_based,
)
from pm4py.algo.evaluation.precision.variants import (
    etconformance_token as precision_token_based,
)
from pm4py.algo.evaluation.replay_fitness.variants import (
    token_replay as fitness_token_based,
)
from pm4py.algo.evaluation.simplicity.variants import (
    arc_degree as simplicity_arc_degree,
)
from pm4py.objects import log as log_lib
from pm4py.objects.conversion.log import converter as log_conversion
from pm4py.util import constants
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    PARAM_FITNESS_WEIGHT = "fitness_weight"
    PARAM_PRECISION_WEIGHT = "precision_weight"
    PARAM_SIMPLICITY_WEIGHT = "simplicity_weight"
    PARAM_GENERALIZATION_WEIGHT = "generalization_weight"


def apply(
    log: Union[EventLog, pd.DataFrame],
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, float]:
    """
    Calculates all metrics based on token-based replay and returns a unified dictionary

    Parameters
    -----------
    log
        Log
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters

    Returns
    -----------
    dictionary
        Dictionary containing fitness, precision, generalization and simplicity; along with the average weight of
        these metrics
    """
    if parameters is None:
        parameters = {}
    log = log_conversion.apply(log, parameters, log_conversion.TO_EVENT_LOG)

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, log_lib.util.xes.DEFAULT_NAME_KEY
    )
    fitness_weight = exec_utils.get_param_value(
        Parameters.PARAM_FITNESS_WEIGHT, parameters, 0.25
    )
    precision_weight = exec_utils.get_param_value(
        Parameters.PARAM_PRECISION_WEIGHT, parameters, 0.25
    )
    simplicity_weight = exec_utils.get_param_value(
        Parameters.PARAM_SIMPLICITY_WEIGHT, parameters, 0.25
    )
    generalization_weight = exec_utils.get_param_value(
        Parameters.PARAM_GENERALIZATION_WEIGHT, parameters, 0.25
    )

    sum_of_weights = (
        fitness_weight
        + precision_weight
        + simplicity_weight
        + generalization_weight
    )
    fitness_weight = fitness_weight / sum_of_weights
    precision_weight = precision_weight / sum_of_weights
    simplicity_weight = simplicity_weight / sum_of_weights
    generalization_weight = generalization_weight / sum_of_weights

    parameters_tr = {token_replay.Parameters.ACTIVITY_KEY: activity_key}

    aligned_traces = token_replay.apply(
        log, net, initial_marking, final_marking, parameters=parameters_tr
    )

    parameters = {token_replay.Parameters.ACTIVITY_KEY: activity_key}

    fitness = fitness_token_based.evaluate(aligned_traces)
    precision = precision_token_based.apply(
        log, net, initial_marking, final_marking, parameters=parameters
    )
    generalization = generalization_token_based.get_generalization(
        net, aligned_traces
    )
    simplicity = simplicity_arc_degree.apply(net)

    metrics_average_weight = (
        fitness_weight * fitness["log_fitness"]
        + precision_weight * precision
        + generalization_weight * generalization
        + simplicity_weight * simplicity
    )

    fscore = 0.0
    if (fitness["log_fitness"] + precision) > 0:
        fscore = (2 * fitness["log_fitness"] * precision) / (
            fitness["log_fitness"] + precision
        )
    dictionary = {
        "fitness": fitness,
        "precision": precision,
        "generalization": generalization,
        "simplicity": simplicity,
        "metricsAverageWeight": metrics_average_weight,
        "fscore": fscore,
    }

    return dictionary



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from sklearn.tree import DecisionTreeClassifier
from sklearn.tree import export_text
from typing import Dict, Tuple, Set, List


def apply(
    clf: DecisionTreeClassifier, columns: List[str]
) -> Tuple[Dict[str, str], Dict[str, Set[str]]]:
    """
    Translates a decision tree object into a dictionary
    associating a set of conditions for each target class

    Parameters
    ----------------
    clf
        Decision tree classifier
    columns
        Columns

    Returns
    ----------------
    dict_classes
        Dictionary associating a set of conditions for each target class
    """
    tree_string = export_text(clf).split("\n")
    levels = {}
    target_classes = {}
    variables = {}
    i = 0
    while i < len(tree_string):
        if "---" in tree_string[i]:
            level = len(tree_string[i].split("|")) - 2
            this_part = tree_string[i].split("--- ")[1]
            this_part_idx_space = this_part.index(" ")
            this_part_0 = this_part[:this_part_idx_space]
            this_part_1 = this_part[this_part_idx_space + 1:]
            if "class" in this_part:
                all_levels = (
                    "(" + " && ".join([levels[i] for i in range(level)]) + ")"
                )
                target_class = this_part.split(": ")[-1]
                if target_class not in target_classes:
                    target_classes[target_class] = []
                target_classes[target_class].append(all_levels)
                if target_class not in variables:
                    variables[target_class] = set()
                for j in range(level):
                    variables[target_class].add(levels[j].split(" ")[0])
            else:
                levels[level] = (
                    columns[int(this_part_0.split("_")[1])] + " " + this_part_1
                )
        i = i + 1
    for cl in target_classes:
        target_classes[cl] = " || ".join(target_classes[cl])
    return target_classes, variables



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.transition_system import obj


def add_arc_from_to(name, fr, to, ts, data=None):
    """
    Adds a transition from a state to another state in some transition system.
    Assumes from and to are in the transition system!

    Parameters
    ----------
    name: name of the transition
    fr: state from
    to:  state to
    ts: transition system to use
    data: data associated to the Transition System

    Returns
    -------
    None
    """
    tran = obj.TransitionSystem.Transition(name, fr, to, data)
    ts.transitions.add(tran)
    fr.outgoing.add(tran)
    to.incoming.add(tran)


def remove_arc_from_to(name, fr, to, ts):
    """
    Removes a transition with a specific name from a state to another state in some transition system.
    Assumes from and to are in the transition system!

    Parameters
    ----------
    name: name of the transition
    fr: state from
    to:  state to
    ts: transition system to use

    Returns
    -------
    None
    """
    ts.transitions = [t for t in ts.transitions if t.name != name]
    fr.outgoing = [t for t in fr.outgoing if t.name != name]
    to.incoming = [t for t in to.incoming if t.name != name]


def remove_all_arcs_from_to(fr, to, ts):
    """
    Removes all transitions from a state to another state in some transition system.
    Assumes from and to are in the transition system!

    Parameters
    ----------
    fr: state from
    to:  state to
    ts: transition system to use

    Returns
    -------
    None
    """
    names_transitions_to_delete = [
        t.name for t in ts.transitions if t in fr.outgoing and t in to.incoming
    ]
    ts.transitions = set(
        [
            t
            for t in ts.transitions
            if t.name not in names_transitions_to_delete
        ]
    )
    fr.outgoing = set(
        [t for t in fr.outgoing if t.name not in names_transitions_to_delete]
    )
    to.incoming = set(
        [t for t in to.incoming if t.name not in names_transitions_to_delete]
    )


def transitive_reduction(ts):
    """
    Computes the transitive reduction of an acyclic transition system.
    Assumes the transition system in input to be acyclic.

    Parameters
    ----------
    ts: acyclic transition system to use

    Returns
    -------
    None
    """

    def check(this_state, this_child, this_done):
        if this_child not in this_done:
            child_children = [
                tr.to_state
                for tr in ts.transitions
                if tr in this_child.outgoing
            ]
            for child_child in child_children:
                remove_all_arcs_from_to(this_state, child_child, ts)
                check(this_state, child_child, this_done)
            this_done.add(this_child)

    for state in ts.states:
        done = set()
        children = [
            tr.to_state for tr in ts.transitions if tr in state.outgoing
        ]
        for child in children:
            check(state, child, done)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import pm4py
import sys
import os
import itertools
from pathlib import Path
import traceback
import pandas as pd
from pm4py.util import constants, pandas_utils


methods = {
    "ConvertToXES": {
        "inputs": [".csv"],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.convert_to_event_log(
                pm4py.format_dataframe(pandas_utils.read_csv(x[0]))
            ),
            x[1],
        ),
    },
    "ConvertToCSV": {
        "inputs": [".xes"],
        "output_extension": ".csv",
        "method": lambda x: pm4py.convert_to_dataframe(
            pm4py.read_xes(x[0])
        ).to_csv(x[1], index=False),
    },
    "ConvertPNMLtoBPMN": {
        "inputs": [".pnml"],
        "output_extension": ".bpmn",
        "method": lambda x: pm4py.write_bpmn(
            pm4py.convert_to_bpmn(*pm4py.read_pnml(x[0])), x[1]
        ),
    },
    "ConvertPNMLtoPTML": {
        "inputs": [".pnml"],
        "output_extension": ".ptml",
        "method": lambda x: pm4py.write_ptml(
            pm4py.convert_to_process_tree(*pm4py.read_pnml(x[0])), x[1]
        ),
    },
    "ConvertPTMLtoPNML": {
        "inputs": [".ptml"],
        "output_extension": ".pnml",
        "method": lambda x: pm4py.write_pnml(
            *pm4py.convert_to_petri_net(pm4py.read_ptml(x[0])), x[1]
        ),
    },
    "ConvertPTMLtoBPMN": {
        "inputs": [".ptml"],
        "output_extension": ".bpmn",
        "method": lambda x: pm4py.write_bpmn(
            pm4py.convert_to_bpmn(pm4py.read_ptml(x[0])), x[1]
        ),
    },
    "ConvertBPMNtoPNML": {
        "inputs": [".bpmn"],
        "output_extension": ".pnml",
        "method": lambda x: pm4py.write_pnml(
            *pm4py.convert_to_petri_net(pm4py.read_bpmn(x[0])), x[1]
        ),
    },
    "ConvertDFGtoPNML": {
        "inputs": [".dfg"],
        "output_extension": ".pnml",
        "method": lambda x: pm4py.write_pnml(
            *pm4py.convert_to_petri_net(*pm4py.read_dfg(x[0])), x[1]
        ),
    },
    "DiscoverPetriNetAlpha": {
        "inputs": [".xes"],
        "output_extension": ".pnml",
        "method": lambda x: pm4py.write_pnml(
            *pm4py.discover_petri_net_alpha(__read_log(x[0])), x[1]
        ),
    },
    "DiscoverPetriNetInductive": {
        "inputs": [".xes"],
        "output_extension": ".pnml",
        "method": lambda x: pm4py.write_pnml(
            *pm4py.discover_petri_net_inductive(__read_log(x[0])), x[1]
        ),
    },
    "DiscoverPetriNetHeuristics": {
        "inputs": [".xes"],
        "output_extension": ".pnml",
        "method": lambda x: pm4py.write_pnml(
            *pm4py.discover_petri_net_heuristics(__read_log(x[0])), x[1]
        ),
    },
    "DiscoverBPMNInductive": {
        "inputs": [".xes"],
        "output_extension": ".bpmn",
        "method": lambda x: pm4py.write_bpmn(
            pm4py.discover_bpmn_inductive(__read_log(x[0])), x[1]
        ),
    },
    "DiscoverProcessTreeInductive": {
        "inputs": [".xes"],
        "output_extension": ".ptml",
        "method": lambda x: pm4py.write_ptml(
            pm4py.discover_process_tree_inductive(__read_log(x[0])), x[1]
        ),
    },
    "DiscoverDFG": {
        "inputs": [".xes"],
        "output_extension": ".dfg",
        "method": lambda x: pm4py.write_dfg(
            *pm4py.discover_dfg(__read_log(x[0])), x[1]
        ),
    },
    "ConformanceDiagnosticsTBR": {
        "inputs": [".xes", ".pnml"],
        "output_extension": ".txt",
        "method": lambda x: open(x[2], "w").write(
            str(
                pm4py.conformance_diagnostics_token_based_replay(
                    __read_log(x[0]),
                    *pm4py.read_pnml(x[1]),
                    return_diagnostics_dataframe=False
                )
            )
        ),
    },
    "ConformanceDiagnosticsAlignments": {
        "inputs": [".xes", ".pnml"],
        "output_extension": ".txt",
        "method": lambda x: open(x[2], "w").write(
            str(
                pm4py.conformance_diagnostics_alignments(
                    __read_log(x[0]),
                    *pm4py.read_pnml(x[1]),
                    return_diagnostics_dataframe=False
                )
            )
        ),
    },
    "FitnessTBR": {
        "inputs": [".xes", ".pnml"],
        "output_extension": ".txt",
        "method": lambda x: open(x[2], "w").write(
            str(
                pm4py.fitness_token_based_replay(
                    __read_log(x[0]), *pm4py.read_pnml(x[1])
                )
            )
        ),
    },
    "FitnessAlignments": {
        "inputs": [".xes", ".pnml"],
        "output_extension": ".txt",
        "method": lambda x: open(x[2], "w").write(
            str(
                pm4py.fitness_alignments(
                    __read_log(x[0]), *pm4py.read_pnml(x[1])
                )
            )
        ),
    },
    "PrecisionTBR": {
        "inputs": [".xes", ".pnml"],
        "output_extension": ".txt",
        "method": lambda x: open(x[2], "w").write(
            str(
                pm4py.precision_token_based_replay(
                    __read_log(x[0]), *pm4py.read_pnml(x[1])
                )
            )
        ),
    },
    "PrecisionAlignments": {
        "inputs": [".xes", ".pnml"],
        "output_extension": ".txt",
        "method": lambda x: open(x[2], "w").write(
            str(
                pm4py.precision_alignments(
                    __read_log(x[0]), *pm4py.read_pnml(x[1])
                )
            )
        ),
    },
    "SaveVisDFG": {
        "inputs": [".dfg"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_dfg(*pm4py.read_dfg(x[0]), x[1]),
    },
    "SaveVisPNML": {
        "inputs": [".pnml"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_petri_net(
            *pm4py.read_pnml(x[0]), x[1]
        ),
    },
    "SaveVisBPMN": {
        "inputs": [".bpmn"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_bpmn(pm4py.read_bpmn(x[0]), x[1]),
    },
    "SaveVisPTML": {
        "inputs": [".ptml"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_process_tree(
            pm4py.read_ptml(x[0]), x[1]
        ),
    },
    "SaveVisDottedChart": {
        "inputs": [".xes", None, None, None],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_dotted_chart(
            __read_log(x[0]), x[4], attributes=[x[1], x[2], x[3]]
        ),
    },
    "SaveVisTransitionSystem": {
        "inputs": [".xes"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_transition_system(
            pm4py.discover_transition_system(__read_log(x[0])), x[1]
        ),
    },
    "SaveVisTrie": {
        "inputs": [".xes"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_prefix_tree(
            pm4py.discover_prefix_tree(__read_log(x[0])), x[1]
        ),
    },
    "SaveVisEventsDistribution": {
        "inputs": [".xes", None],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_events_distribution_graph(
            __read_log(x[0]), x[2], distr_type=x[1]
        ),
    },
    "SaveVisEventsPerTime": {
        "inputs": [".xes"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_events_per_time_graph(
            __read_log(x[0]), x[1]
        ),
    },
    "GenerateProcessTree": {
        "inputs": [None],
        "output_extension": ".ptml",
        "method": lambda x: pm4py.write_ptml(
            pm4py.generate_process_tree(
                parameters={
                    "min": int(x[0]),
                    "max": int(x[0]),
                    "mode": int(x[0]),
                }
            ),
            x[1],
        ),
    },
    "PNMLplayout": {
        "inputs": [".pnml"],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.play_out(*pm4py.read_pnml(x[0])), x[1]
        ),
    },
    "PTMLplayout": {
        "inputs": [".ptml"],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.play_out(pm4py.read_ptml(x[0])), x[1]
        ),
    },
    "DFGplayout": {
        "inputs": [".dfg"],
        "output_extension": ".dfg",
        "method": lambda x: pm4py.write_xes(
            pm4py.play_out(*pm4py.read_dfg(x[0])), x[1]
        ),
    },
    "SaveVisSNA": {
        "inputs": [".xes", None],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_sna(
            __apply_sna(__read_log(x[0]), x[1]), x[2]
        ),
    },
    "SaveVisCaseDuration": {
        "inputs": [".xes"],
        "output_extension": ".png",
        "method": lambda x: pm4py.save_vis_case_duration_graph(
            __read_log(x[0]), x[1]
        ),
    },
    "FilterVariantsTopK": {
        "inputs": [".xes", None],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.filter_variants_top_k(__read_log(x[0]), int(x[1])), x[2]
        ),
    },
    "FilterVariantsCoverage": {
        "inputs": [".xes", None],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.filter_variants_by_coverage_percentage(
                __read_log(x[0]), float(x[1])
            ),
            x[2],
        ),
    },
    "FilterCasePerformance": {
        "inputs": [".xes", None, None],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.filter_case_performance(
                __read_log(x[0]),
                min_performance=float(x[1]),
                max_performance=float(x[2]),
            ),
            x[3],
        ),
    },
    "FilterTimeRange": {
        "inputs": [".xes", None, None],
        "output_extension": ".xes",
        "method": lambda x: pm4py.write_xes(
            pm4py.filter_time_range(
                __read_log(x[0]), x[1] + " 00:00:00", x[2] + " 23:59:59"
            ),
            x[3],
        ),
    },
}


def __read_log(log_path):
    if "xes" in log_path.lower():
        return pm4py.read_xes(log_path)
    elif "csv" in log_path.lower():
        dataframe = pandas_utils.read_csv(log_path)
        dataframe = pm4py.format_dataframe(dataframe)
        return dataframe


def __apply_sna(log, method, **kwargs):
    if method == "handover":
        return pm4py.discover_handover_of_work_network(log, **kwargs)
    elif method == "working_together":
        return pm4py.discover_working_together_network(log, **kwargs)
    elif method == "similar_activities":
        return pm4py.discover_activity_based_resource_similarity(log, **kwargs)
    elif method == "subcontracting":
        return pm4py.discover_subcontracting_network(log, **kwargs)


def __get_output_name(inp_list, idx, method_name, extension):
    ret = []
    for inp in inp_list:
        ret.append(str(Path(inp).stem))
    return "_".join(ret) + "_" + method_name + extension


def cli_interface():
    method_name = sys.argv[1]
    if method_name in methods:
        method = methods[method_name]
        inputs = []
        for i in range(len(method["inputs"])):
            ci = sys.argv[2 + i]
            if os.path.isdir(ci):
                if not os.path.exists(ci):
                    raise Exception(
                        "the provided path (" + ci + ") does not exist."
                    )
                files = os.listdir(ci)
                inputs.append(
                    [
                        os.path.join(ci, f)
                        for f in files
                        if os.path.isfile(os.path.join(ci, f))
                        and f.endswith(method["inputs"][i])
                    ]
                )
            else:
                inputs.append([ci])
        j = 2 + len(method["inputs"])
        inputs = list(itertools.product(*inputs))
        if len(inputs) == 1:
            outputs = [[sys.argv[j]]]
        else:
            if not os.path.exists(sys.argv[j]):
                os.mkdir(sys.argv[j])
            outputs = [
                [
                    os.path.join(
                        sys.argv[j],
                        __get_output_name(
                            inputs[z],
                            z,
                            method_name,
                            method["output_extension"],
                        ),
                    )
                ]
                for z in range(len(inputs))
            ]
        method_tuples = [(*inputs[i], *outputs[i]) for i in range(len(inputs))]
        for i in range(len(method_tuples)):
            try:
                if not os.path.exists(method_tuples[i][-1]):
                    print(method_name, method_tuples[i])
                    method["method"](method_tuples[i])
            except BaseException:
                traceback.print_exc()
    else:
        raise Exception(
            "the provided method ("
            + method_name
            + ") does not exist in the CLI."
        )


if __name__ == "__main__":
    cli_interface()



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import json
import logging
import importlib.util

from pm4py.util.points_subset import pick_chosen_points_list
from pm4py.util import exec_utils, pandas_utils, constants
from enum import Enum


class Parameters(Enum):
    GRAPH_POINTS = "graph_points"
    POINT_TO_SAMPLE = "points_to_sample"


def get_sorted_attributes_list(attributes):
    """
    Gets sorted attributes list

    Parameters
    ----------
    attributes
        Dictionary of attributes associated with their count

    Returns
    ----------
    listact
        Sorted end attributes list
    """
    listattr = []
    for a in attributes:
        listattr.append([a, attributes[a]])
    listattr = sorted(listattr, key=lambda x: x[1], reverse=True)
    return listattr


def get_attributes_threshold(
    alist, decreasing_factor, min_activity_count=1, max_activity_count=25
):
    """
    Get attributes cutting threshold

    Parameters
    ----------
    alist
        Sorted attributes list
    decreasing_factor
        Decreasing factor of the algorithm
    min_activity_count
        Minimum number of activities to include
    max_activity_count
        Maximum number of activities to include

    Returns
    ---------
    threshold
        Activities cutting threshold
    """
    index = max(0, min(min_activity_count - 1, len(alist) - 1))
    threshold = alist[index][1]
    index = index + 1
    for i in range(index, len(alist)):
        value = alist[i][1]
        if value > threshold * decreasing_factor:
            threshold = value
        if i >= max_activity_count:
            break
    return threshold


def get_kde_numeric_attribute(values, parameters=None):
    """
    Gets the KDE estimation for the distribution of a numeric attribute values

    Parameters
    -------------
    values
        Values of the numeric attribute value
    parameters
        Possible parameters of the algorithm, including:
            graph_points -> number of points to include in the graph

    Returns
    --------------
    x
        X-axis values to represent (including the exact min and max)
    y
        Y-axis values to represent
    """
    if importlib.util.find_spec("scipy") and importlib.util.find_spec("numpy"):
        from scipy.stats import gaussian_kde
        import numpy as np

        if parameters is None:
            parameters = {}

        graph_points = exec_utils.get_param_value(
            Parameters.GRAPH_POINTS, parameters, 200
        )
        values = np.sort(values)
        
        # Check if we have enough unique values for KDE
        unique_values = np.unique(values)
        if len(unique_values) < 2:
            # Handle edge case: not enough unique values for KDE
            if len(unique_values) == 0:
                # No values at all
                return [], []
            else:
                # Single unique value - create a simple representation
                single_val = float(unique_values[0])
                # Create a small range around the single value for visualization
                eps = max(abs(single_val) * 0.01, 1e-6) if single_val != 0 else 1.0
                xs = np.linspace(single_val - eps, single_val + eps, graph_points)
                # Create a spike at the single value
                ys = np.zeros(graph_points)
                mid_idx = graph_points // 2
                ys[mid_idx] = 1.0
                return xs.tolist(), ys.tolist()
        
        density = gaussian_kde(values)

        # ensure we have at least two points for each spacing
        half = max(int(graph_points // 2), 2)

        min_val, max_val = values[0], values[-1]
        eps = 1e-6

        # linear space including both endpoints
        xs1 = np.linspace(min_val, max_val, half, endpoint=True)
        # try to enrich the sampling near the distribution tails; fall back when geomspace is not applicable
        if min_val > 0 and max_val > 0:
            # both bounds positive -> standard geometric spacing
            xs2 = np.geomspace(max(min_val, eps), max_val, half, endpoint=True)
        elif min_val < 0 and max_val < 0:
            # both bounds negative -> mirror geometric spacing on the absolute values
            xs2 = -np.geomspace(abs(min_val), max(abs(max_val), eps), half, endpoint=True)
        else:
            # bounds cross or hit zero -> stick to linear spacing to avoid invalid geometric ranges
            xs2 = np.linspace(min_val, max_val, half, endpoint=True)

        # combine, add exact endpoints, dedupe & sort
        xs = np.unique(
            np.concatenate([xs1, xs2, [min_val, max_val]])
        )

        return xs.tolist(), density(xs).tolist()
    else:
        msg = "scipy is not available. graphs cannot be built!"
        logging.error(msg)
        raise Exception(msg)


def get_kde_numeric_attribute_json(values, parameters=None):
    """
    Gets the KDE estimation for the distribution of a numeric attribute values
    (expressed as JSON)

    Parameters
    --------------
    values
        Values of the numeric attribute value
    parameters
        Possible parameters of the algorithm, including:
            graph_points: number of points to include in the graph

    Returns
    --------------
    json
        JSON representing the graph points
    """
    x, y = get_kde_numeric_attribute(values, parameters=parameters)

    ret = []
    for i in range(len(x)):
        ret.append((x[i], y[i]))

    return json.dumps(ret)


def get_kde_date_attribute(values, parameters=None):
    """
    Gets the KDE estimation for the distribution of a date attribute values

    Parameters
    -------------
    values
        Values of the date attribute value
    parameters
        Possible parameters of the algorithm, including:
            graph_points -> number of points to include in the graph


    Returns
    --------------
    x
        X-axis values to represent
    y
        Y-axis values to represent
    """
    if importlib.util.find_spec("scipy") and importlib.util.find_spec("numpy"):
        from scipy.stats import gaussian_kde
        import numpy as np
        import pandas as pd

        if parameters is None:
            parameters = {}

        graph_points = exec_utils.get_param_value(
            Parameters.GRAPH_POINTS, parameters, 200
        )
        points_to_sample = exec_utils.get_param_value(
            Parameters.POINT_TO_SAMPLE, parameters, 400
        )

        red_values = pick_chosen_points_list(points_to_sample, values, include_extremes=True)
        int_values = sorted(
            [x.replace(tzinfo=None).timestamp() for x in red_values]
        )
        
        # Check if we have enough unique values for KDE
        unique_int_values = np.unique(int_values)
        if len(unique_int_values) < 2:
            # Handle edge case: not enough unique values for KDE
            if len(unique_int_values) == 0:
                # No values at all
                return [[], []]
            else:
                # Single unique value - create a simple representation
                single_val = float(unique_int_values[0])
                # Create a small time range around the single value (1 hour range)
                time_eps = 3600  # 1 hour in seconds
                xs = np.linspace(single_val - time_eps, single_val + time_eps, graph_points)
                xs_transf = pd.to_datetime(xs * 10**9, unit="ns")
                # Create a spike at the single value
                ys = np.zeros(graph_points)
                mid_idx = graph_points // 2
                ys[mid_idx] = 1.0
                return [xs_transf, ys.tolist()]
        
        density = gaussian_kde(int_values)
        xs = np.linspace(min(int_values), max(int_values), graph_points)
        xs_transf = pd.to_datetime(xs * 10**9, unit="ns")

        return [xs_transf, density(xs)]
    else:
        msg = "scipy is not available. graphs cannot be built!"
        logging.error(msg)
        raise Exception(msg)


def get_kde_date_attribute_json(values, parameters=None):
    """
    Gets the KDE estimation for the distribution of a date attribute values
    (expressed as JSON)

    Parameters
    --------------
    values
        Values of the date attribute value
    parameters
        Possible parameters of the algorithm, including:
            graph_points: number of points to include in the graph

    Returns
    --------------
    json
        JSON representing the graph points
    """
    x, y = get_kde_date_attribute(values, parameters=parameters)

    ret = []
    for i in range(len(x)):
        ret.append((x[i].replace(tzinfo=None).timestamp(), y[i]))

    return json.dumps(ret)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
import polars as pl
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.util.business_hours import soj_time_business_hours_diff
from typing import Optional, Dict, Any, Union, Tuple


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    KEEP_FIRST_FOLLOWING = "keep_first_following"


def get_partial_order_dataframe(
    lf: pl.LazyFrame,
    start_timestamp_key=None,
    timestamp_key="time:timestamp",
    case_id_glue="case:concept:name",
    activity_key="concept:name",
    keep_first_following=True,
    business_hours=False,
    business_hours_slot=None,
    workcalendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
) -> pl.LazyFrame:
    """
    Gets the partial order between events (of the same case) in a Polars LazyFrame

    Parameters
    --------------
    lf
        LazyFrame
    start_timestamp_key
        Start timestamp key (if not provided, defaulted to the timestamp_key)
    timestamp_key
        Complete timestamp
    case_id_glue
        Column of the dataframe to use as case ID
    activity_key
        Activity key
    keep_first_following
        Keep only the first event following the given event
    business_hours
        Whether to compute business hours
    business_hours_slot
        Business hours slot configuration
    workcalendar
        Work calendar configuration

    Returns
    ---------------
    part_ord_dataframe
        Partial order LazyFrame (with flow time between events)
    """
    # if not differently specified, set the start timestamp key to the timestamp key
    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY
        lf = lf.with_columns(pl.col(timestamp_key).alias(start_timestamp_key))

    # Reduce to needed columns for efficiency
    needed_columns = [case_id_glue, activity_key, start_timestamp_key, timestamp_key]
    lf = lf.select(needed_columns)

    # Sort by case ID and timestamps
    lf = lf.sort([case_id_glue, start_timestamp_key, timestamp_key])
    
    # Add index for ordering within cases
    lf = lf.with_row_count("__index__")
    
    # Self-join to get all pairs within same cases
    lf_self = lf.select([
        pl.col(case_id_glue).alias(case_id_glue + "_2"),
        pl.col(activity_key).alias(activity_key + "_2"),
        pl.col(start_timestamp_key).alias(start_timestamp_key + "_2"),
        pl.col(timestamp_key).alias(timestamp_key + "_2"),
        pl.col("__index__").alias("__index___2")
    ])
    
    # Join on case ID
    partial_order_df = lf.join(
        lf_self, 
        left_on=case_id_glue, 
        right_on=case_id_glue + "_2",
        how="inner"
    )
    
    # Filter to ensure proper ordering and temporal relationships
    partial_order_df = partial_order_df.filter(
        (pl.col("__index__") < pl.col("__index___2")) &
        (pl.col(timestamp_key) <= pl.col(start_timestamp_key + "_2"))
    )
    
    # Calculate flow time
    if business_hours and business_hours_slot is not None:
        # For business hours, we need to collect and apply Python function
        # This is less efficient but required for business hours calculation
        df_collected = partial_order_df.collect()
        flow_times = []
        for row in df_collected.iter_rows():
            ts_idx = df_collected.columns.index(timestamp_key)
            start_ts_2_idx = df_collected.columns.index(start_timestamp_key + "_2")
            flow_time = soj_time_business_hours_diff(
                row[ts_idx], row[start_ts_2_idx], business_hours_slot, workcalendar
            )
            flow_times.append(flow_time)
        
        df_collected = df_collected.with_columns(
            pl.Series(constants.DEFAULT_FLOW_TIME, flow_times)
        )
        partial_order_df = df_collected.lazy()
    else:
        # Standard time difference calculation
        partial_order_df = partial_order_df.with_columns(
            (pl.col(start_timestamp_key + "_2") - pl.col(timestamp_key))
            .dt.total_seconds()
            .alias(constants.DEFAULT_FLOW_TIME)
        )
    
    # Keep only first following if specified
    if keep_first_following:
        partial_order_df = partial_order_df.group_by("__index__").first()
    
    return partial_order_df


def apply(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[Tuple[str, str], int]:
    """
    Computes the eventually-follows relationships between activities

    Parameters
    --------------
    lf
        Polars LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    --------------
    ret_dict
        Dictionary with eventually-follows relationships and their counts
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    keep_first_following = exec_utils.get_param_value(
        Parameters.KEEP_FIRST_FOLLOWING, parameters, False
    )

    partial_order_dataframe = get_partial_order_dataframe(
        lf,
        start_timestamp_key=start_timestamp_key,
        timestamp_key=timestamp_key,
        case_id_glue=case_id_glue,
        activity_key=activity_key,
        keep_first_following=keep_first_following,
    )

    # Group by activity pairs and count
    ret_dict_df = partial_order_dataframe.group_by([activity_key, activity_key + "_2"]).count().collect()
    
    ret_dict = {}
    for row in ret_dict_df.iter_rows():
        key = (row[0], row[1])
        count = row[2]
        ret_dict[key] = int(count)

    return ret_dict


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.objects.log.obj import Trace
from typing import Optional, Dict, Any


class Parameters(Enum):
    INCLUDE_CASE_ATTRIBUTES = "include_case_attributes"
    INCLUDE_EVENT_ATTRIBUTES = "include_event_attributes"
    INCLUDE_TIMESTAMP = "include_timestamp"
    INCLUDE_HEADER = "include_header"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY


def apply(case: Trace, parameters: Optional[Dict[Any, Any]] = None) -> str:
    """
    Provides a textual abstraction of a single case (Trace object) of a traditional event log

    Parameters
    ---------------
    case
        Single case (Trace object) of a traditional event log
    parameters
        Parameters of the method, including:
        - Parameters.INCLUDING_CASE_ATTRIBUTES
        - Parameters.INCLUDE_EVENT_ATTRIBUTES
        - Parameters.INCLUDE_TIMESTAMP
        - Parameters.INCLUDE_HEADER => includes the header (or not) in the response)
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.TIMESTAMP_KEY => the attribute to be used as timestamp

    Returns
    ---------------
    stru
        Textual abstraction of the case
    """
    if parameters is None:
        parameters = {}

    include_case_attributes = exec_utils.get_param_value(
        Parameters.INCLUDE_CASE_ATTRIBUTES, parameters, True
    )
    include_event_attributes = exec_utils.get_param_value(
        Parameters.INCLUDE_EVENT_ATTRIBUTES, parameters, True
    )
    include_timestamp = exec_utils.get_param_value(
        Parameters.INCLUDE_TIMESTAMP, parameters, True
    )
    include_header = exec_utils.get_param_value(
        Parameters.INCLUDE_HEADER, parameters, True
    )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    ret = ["\n"]

    if include_header:
        ret.append("If I have a case")
        if include_case_attributes:
            ret.append(" with the following (case) attributes:")
        ret.append("")

    ret.append("\n")

    if include_case_attributes:
        for k, v in case.attributes.items():
            ret.append("\n%s = %s" % (str(k), str(v)))

    ret.append("\n\n")

    if include_header:
        ret.append(
            "\nthe case contains the following events (the activity of the event is reported first):\n\n"
        )

    for ev in case:
        stru = "%s " % (ev[activity_key])
        ev_attrs = sorted(
            [str(x) for x in ev if x not in [activity_key, timestamp_key]]
        )

        if include_timestamp or (ev_attrs and include_event_attributes):
            stru += "("
            if include_timestamp:
                stru += " timestamp = " + str(ev[timestamp_key]) + " ; "

            if include_event_attributes and ev_attrs:
                for attr in ev_attrs:
                    stru += " " + attr + " = " + str(ev[attr]) + " ; "

            stru += ")"

        stru += "\n"

        ret.append(stru)

    ret.append("\n")

    return "".join(ret)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects import log as log_lib
from pm4py.algo.evaluation.precision import utils as precision_utils
from pm4py.objects.petri_net.utils import align_utils as utils, check_soundness
from pm4py.objects.petri_net.utils.petri_utils import construct_trace_net
from pm4py.objects.petri_net.utils.synchronous_product import construct
from pm4py.statistics.start_activities.log.get import get_start_activities
from pm4py.objects.petri_net.utils.align_utils import (
    get_visible_transitions_eventually_enabled_by_marking,
)
from pm4py.util import exec_utils
from pm4py.util import xes_constants, pandas_utils, thread_utils
from pm4py.utils import is_polars_lazyframe
import importlib.util
from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.conversion.log import converter as log_converter
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TOKEN_REPLAY_VARIANT = "token_replay_variant"
    CLEANING_TOKEN_FLOOD = "cleaning_token_flood"
    SHOW_PROGRESS_BAR = "show_progress_bar"
    MULTIPROCESSING = "multiprocessing"
    CORES = "cores"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    net: PetriNet,
    marking: Marking,
    final_marking: Marking,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Get Align-ET Conformance precision

    Parameters
    ----------
    log
        Trace log
    net
        Petri net
    marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Activity key
    """

    if parameters is None:
        parameters = {}

    debug_level = (
        parameters["debug_level"] if "debug_level" in parameters else 0
    )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, log_lib.util.xes.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    # default value for precision, when no activated transitions (not even by
    # looking at the initial marking) are found
    precision = 1.0
    sum_ee = 0
    sum_at = 0
    unfit = 0

    if not check_soundness.check_easy_soundness_net_in_fin_marking(
        net, marking, final_marking
    ):
        raise Exception(
            "trying to apply Align-ETConformance on a Petri net that is not a easy sound net!!"
        )

    if not pandas_utils.check_is_pandas_dataframe(log):
        log = log_converter.apply(
            log,
            variant=log_converter.Variants.TO_EVENT_LOG,
            parameters=parameters,
        )

    prefixes, prefix_count = precision_utils.get_log_prefixes(
        log, activity_key=activity_key, case_id_key=case_id_key
    )
    prefixes_keys = list(prefixes.keys())
    fake_log = precision_utils.form_fake_log(
        prefixes_keys, activity_key=activity_key
    )

    align_stop_marking = align_fake_log_stop_marking(
        fake_log, net, marking, final_marking, parameters=parameters
    )
    all_markings = transform_markings_from_sync_to_original_net(
        align_stop_marking, net, parameters=parameters
    )

    for i in range(len(prefixes)):
        markings = all_markings[i]

        if markings is not None:
            log_transitions = set(prefixes[prefixes_keys[i]])
            activated_transitions_labels = set()
            for m in markings:
                # add to the set of activated transitions in the model the activated transitions
                # for each prefix
                activated_transitions_labels = activated_transitions_labels.union(
                    x.label
                    for x in utils.get_visible_transitions_eventually_enabled_by_marking(
                        net, m
                    )
                    if x.label is not None
                )
            escaping_edges = activated_transitions_labels.difference(
                log_transitions
            )

            sum_at += (
                len(activated_transitions_labels)
                * prefix_count[prefixes_keys[i]]
            )
            sum_ee += len(escaping_edges) * prefix_count[prefixes_keys[i]]

            if debug_level > 1:
                print("")
                print("prefix=", prefixes_keys[i])
                print("log_transitions=", log_transitions)
                print("activated_transitions=", activated_transitions_labels)
                print("escaping_edges=", escaping_edges)
        else:
            unfit += prefix_count[prefixes_keys[i]]

    if debug_level > 0:
        print("\n")
        print("overall unfit", unfit)
        print("overall activated transitions", sum_at)
        print("overall escaping edges", sum_ee)

    # fix: also the empty prefix should be counted!
    start_activities = set(get_start_activities(log, parameters=parameters))
    trans_en_ini_marking = set(
        [
            x.label
            for x in get_visible_transitions_eventually_enabled_by_marking(
                net, marking
            )
        ]
    )
    diff = trans_en_ini_marking.difference(start_activities)
    if type(log) is EventLog:
        n_traces = len(log)
    elif is_polars_lazyframe(log):
        import polars as pl
        n_traces_result = log.select(pl.col(case_id_key).n_unique()).collect()
        n_traces = 0
        if n_traces_result.height > 0 and n_traces_result.width > 0:
            n_traces = int(n_traces_result.to_series(0)[0] or 0)
    else:
        n_traces = int(log[case_id_key].nunique())
    sum_at += n_traces * len(trans_en_ini_marking)
    sum_ee += n_traces * len(diff)
    # end fix

    if sum_at > 0:
        precision = 1 - float(sum_ee) / float(sum_at)

    return precision


def transform_markings_from_sync_to_original_net(
    markings0, net, parameters=None
):
    """
    Transform the markings of the sync net (in which alignment stops) into markings of the original net
    (in order to measure the precision)

    Parameters
    -------------
    markings0
        Markings on the sync net (expressed as place name with count)
    net
        Petri net
    parameters
        Parameters of the algorithm

    Returns
    -------------
    markings
        Markings of the original model (expressed as place with count)
    """
    if parameters is None:
        parameters = {}

    places_corr = {p.name: p for p in net.places}

    markings = []

    for i in range(len(markings0)):
        res_list = markings0[i]

        # res_list shall be a list of markings.
        # If it is None, then there is no correspondence markings
        # in the original Petri net
        if res_list is not None:
            # saves all the markings reached by the optimal alignment
            # as markings of the original net
            markings.append([])

            for j in range(len(res_list)):
                res = res_list[j]

                atm = Marking()
                for pl, count in res.items():
                    if pl[0] == utils.SKIP:
                        atm[places_corr[pl[1]]] = count
                markings[-1].append(atm)
        else:
            markings.append(None)

    return markings


def align_fake_log_stop_marking(
    fake_log, net, marking, final_marking, parameters=None
):
    """
    Align the 'fake' log with all the prefixes in order to get the markings in which
    the alignment stops

    Parameters
    -------------
    fake_log
        Fake log
    net
        Petri net
    marking
        Marking
    final_marking
        Final marking
    parameters
        Parameters of the algorithm

    Returns
    -------------
    alignment
        For each trace in the log, return the marking in which the alignment stops (expressed as place name with count)
    """
    if parameters is None:
        parameters = {}

    show_progress_bar = exec_utils.get_param_value(
        Parameters.SHOW_PROGRESS_BAR, parameters, constants.SHOW_PROGRESS_BAR
    )
    multiprocessing = exec_utils.get_param_value(
        Parameters.MULTIPROCESSING,
        parameters,
        constants.ENABLE_MULTIPROCESSING_DEFAULT,
    )

    progress = None
    if (
        importlib.util.find_spec("tqdm")
        and show_progress_bar
        and len(fake_log) > 1
    ):
        from tqdm.auto import tqdm

        progress = tqdm(
            total=len(fake_log),
            desc="computing precision with alignments, completed variants :: ",
        )

    if multiprocessing:
        align_intermediate_result = (
            __align_log_with_multiprocessing_stop_marking(
                fake_log,
                net,
                marking,
                final_marking,
                progress,
                parameters=parameters,
            )
        )
    else:
        align_intermediate_result = (
            __align_log_wo_multiprocessing_stop_marking(
                fake_log,
                net,
                marking,
                final_marking,
                progress,
                parameters=parameters,
            )
        )

    align_result = []
    for i in range(len(align_intermediate_result)):
        res = align_intermediate_result[i]
        if res is not None:
            align_result.append([])
            for mark in res:
                res2 = {}
                for pl in mark:
                    # transforms the markings for easier correspondence at the end
                    # (distributed engine friendly!)
                    res2[(pl.name[0], pl.name[1])] = mark[pl]

                align_result[-1].append(res2)
        else:
            # if there is no path from the initial marking
            # replaying the given prefix, then add None
            align_result.append(None)

    # gracefully close progress bar
    if progress is not None:
        progress.close()
    del progress

    return align_result


def __align_log_wo_multiprocessing_stop_marking(
    fake_log, net, marking, final_marking, progress, parameters=None
):
    align_intermediate_result = [None] * len(fake_log)

    thm = thread_utils.Pm4pyThreadManager()

    def _compute(idx, trace, results):
        results[idx] = __align_trace_stop_marking(trace, net, marking, final_marking, parameters=parameters)
        if progress is not None:
            progress.update()

    for idx, trace in enumerate(fake_log):
        thm.submit(_compute, idx, trace, align_intermediate_result)

    thm.join()

    return align_intermediate_result


def __align_log_with_multiprocessing_stop_marking(
    fake_log, net, marking, final_marking, progress, parameters=None
):
    if parameters is not None:
        parameters = {}

    import multiprocessing
    from concurrent.futures import ProcessPoolExecutor

    num_cores = exec_utils.get_param_value(
        Parameters.CORES, parameters, multiprocessing.cpu_count() - 2
    )
    align_intermediate_result = []
    with ProcessPoolExecutor(max_workers=num_cores) as executor:
        futures = []
        for i in range(len(fake_log)):
            futures.append(
                executor.submit(
                    __align_trace_stop_marking,
                    fake_log[i],
                    net,
                    marking,
                    final_marking,
                    parameters,
                )
            )
        if progress is not None:
            alignments_ready = 0
            while alignments_ready != len(futures):
                current = 0
                for index, variant in enumerate(futures):
                    current = current + 1 if futures[index].done() else current
                if current > alignments_ready:
                    for i in range(0, current - alignments_ready):
                        progress.update()
                alignments_ready = current
        for index, variant in enumerate(futures):
            align_intermediate_result.append(futures[index].result())

    return align_intermediate_result


def __align_trace_stop_marking(
    trace, net, marking, final_marking, parameters=None
):
    sync_net, sync_initial_marking, sync_final_marking = build_sync_net(
        trace, net, marking, final_marking, parameters=parameters
    )
    stop_marking = Marking()
    for pl, count in sync_final_marking.items():
        if pl.name[1] == utils.SKIP:
            stop_marking[pl] = count
    cost_function = utils.construct_standard_cost_function(
        sync_net, utils.SKIP
    )

    # perform the alignment of the prefix
    res = precision_utils.__search(
        sync_net,
        sync_initial_marking,
        sync_final_marking,
        stop_marking,
        cost_function,
        utils.SKIP,
    )

    return res


def build_sync_net(
    trace, petri_net, initial_marking, final_marking, parameters=None
):
    """
    Build the sync product net between the Petri net and the trace prefix

    Parameters
    ---------------
    trace
        Trace prefix
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    trace_net, trace_im, trace_fm = construct_trace_net(
        trace, activity_key=activity_key
    )

    sync_prod, sync_initial_marking, sync_final_marking = construct(
        trace_net,
        trace_im,
        trace_fm,
        petri_net,
        initial_marking,
        final_marking,
        utils.SKIP,
    )

    return sync_prod, sync_initial_marking, sync_final_marking



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME, PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.statistics.variants.polars import get as variants_get
from typing import Optional, Dict, Any, Union, List
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    POSITIVE = "positive"


def apply(
    df: pl.LazyFrame,
    admitted_variants: List[List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter on variants

    Parameters
    -----------
    df
        LazyFrame
    admitted_variants
        List of variants to include/exclude
    parameters
        Parameters of the algorithm

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(Parameters.POSITIVE, parameters, True)

    # Create variants
    variants_df = (
        df.sort([case_id_glue, "time:timestamp"])
        .group_by(case_id_glue, maintain_order=True)
        .agg(
            pl.col(activity_key)
            .str.concat(constants.DEFAULT_VARIANT_SEP)
            .alias("variant")
        )
    )

    # Convert admitted variants to strings
    variant_strings = [constants.DEFAULT_VARIANT_SEP.join(v) for v in admitted_variants]
    
    # Filter for matching variants
    matching_cases = variants_df.filter(
        pl.col("variant").is_in(variant_strings)
    ).select(case_id_glue)

    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret


def filter_variants_top_k(
    log: pl.LazyFrame,
    k: int,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Keeps the top-k variants of the log

    Parameters
    -------------
    log
        Event log LazyFrame
    k
        Number of variants that should be kept
    parameters
        Parameters

    Returns
    -------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    variants = variants_get.get_variants_count(log, parameters=parameters)
    variant_count = []
    for variant, count in variants.items():
        variant_count.append([variant, count])
    variant_count = sorted(
        variant_count, key=lambda x: (x[1], x[0]), reverse=True
    )
    variant_count = variant_count[: min(k, len(variant_count))]
    variants_to_filter = [x[0] for x in variant_count]

    return apply(log, variants_to_filter, parameters=parameters)


def apply_auto_filter(
    df: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply auto filter on variants

    Parameters
    -----------
    df
        LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, constants.DEFAULT_NAME_KEY
    )
    
    # Get variant counts
    variant_counts = (
        df.sort([case_id_glue, "time:timestamp"])
        .group_by(case_id_glue, maintain_order=True)
        .agg(
            pl.col(activity_key)
            .str.concat(constants.DEFAULT_VARIANT_SEP)
            .alias("variant")
        )
        .group_by("variant")
        .count()
        .sort("count", descending=True)
    )
    
    # Calculate cumulative percentage
    total_cases = variant_counts.select(pl.sum("count")).collect()[0, 0]
    variant_counts = variant_counts.with_columns(
        (pl.col("count").cumsum() / total_cases).alias("cumulative_percentage")
    )
    
    # Keep variants covering 80% of cases
    selected_variants = (
        variant_counts
        .filter(pl.col("cumulative_percentage") <= 0.8)
        .select("variant")
        .collect()["variant"]
        .to_list()
    )
    
    # Apply filter
    return apply_from_variant_list(df, selected_variants, parameters)


def apply_from_variant_list(
    df: pl.LazyFrame,
    variants: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply filter from variant list

    Parameters
    -----------
    df
        LazyFrame
    variants
        List of variants (as strings)
    parameters
        Parameters

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, constants.DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(Parameters.POSITIVE, parameters, True)

    # Create variants
    variants_df = (
        df.sort([case_id_glue, "time:timestamp"])
        .group_by(case_id_glue, maintain_order=True)
        .agg(
            pl.col(activity_key)
            .str.concat(constants.DEFAULT_VARIANT_SEP)
            .alias("variant")
        )
    )
    
    # Filter for matching variants
    matching_cases = variants_df.filter(
        pl.col("variant").is_in(variants)
    ).select(case_id_glue)

    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.obj import EventLog
from enum import Enum
from pm4py.util import exec_utils
from pm4py.algo.simulation.playout.declare.variants import classic


class Variants(Enum):
    CLASSIC = classic


def apply(declare_model, variant=Variants.CLASSIC, parameters=None) -> EventLog:
    """
    Produce a playout EventLog of a given DECLARE model.
    Optional parameters:
      - n_traces (int): number of traces to generate. Default = 1000
      - min_length (int): minimal length of each trace. Default = 3
      - max_length (int): maximal length of each trace. Default = 15
    """
    return exec_utils.get_variant(variant).apply(declare_model, parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME, PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, List
import polars as pl
from pm4py.algo.filtering.common.traces.infix_to_regex import translate_infix_to_regex
import re


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    DECREASING_FACTOR = "decreasingFactor"
    POSITIVE = "positive"


def get_variants_df(
    df: pl.LazyFrame,
    case_id_key: str = CASE_CONCEPT_NAME,
    activity_key: str = xes_constants.DEFAULT_NAME_KEY,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pl.DataFrame:
    """
    Gets variants dataframe from a LazyFrame
    
    Parameters
    ----------
    df
        LazyFrame
    case_id_key
        Case ID key
    activity_key
        Activity key
    parameters
        Possible parameters of the algorithm
    
    Returns
    ----------
    variants_df
        Variants dataframe
    """
    if parameters is None:
        parameters = {}
    
    # Create variants by grouping activities per case
    variants_df = (
        df.sort([case_id_key, xes_constants.DEFAULT_TIMESTAMP_KEY])
        .group_by(case_id_key, maintain_order=True)
        .agg([
            pl.col(activity_key).alias("variant_list"),
            pl.count().alias("count")
        ])
        .with_columns(
            pl.col("variant_list").list.join(constants.DEFAULT_VARIANT_SEP).alias("variant")
        )
        .collect()
    )
    
    return variants_df


def apply(
    df: pl.LazyFrame,
    admitted_traces: List[List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply a filter on traces using regular expressions

    Parameters
    -----------
    df
        LazyFrame
    admitted_traces
        List of admitted traces (to include/exclude). Can contain "..." wildcard for any sequence of activities
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
            variants_df -> If provided, avoid recalculation of the variants dataframe

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    
    # Get or calculate variants dataframe
    variants_df = parameters.get("variants_df")
    if variants_df is None:
        variants_df = get_variants_df(df, case_id_key=case_id_glue, activity_key=activity_key, parameters=parameters)
    
    # Build regex pattern from admitted traces
    filter_regex = "|".join(
        [f"({translate_infix_to_regex(inf)})" for inf in admitted_traces]
    )
    
    # Apply regex matching
    # Since Polars doesn't have a direct apply with lambda for regex, we need to use a different approach
    matching_cases = []
    for idx, row in enumerate(variants_df.iter_rows(named=True)):
        variant = row["variant"]
        if bool(re.search(filter_regex, variant)):
            matching_cases.append(row[case_id_glue])
    
    # Get the data type of the case ID column from the original dataframe
    case_id_dtype = df.select(pl.col(case_id_glue)).dtypes[0]
    
    # Create a DataFrame with matching cases
    if matching_cases:
        matching_df = pl.DataFrame({case_id_glue: matching_cases}).lazy()
    else:
        # Create empty DataFrame with the correct schema
        matching_df = pl.DataFrame({case_id_glue: []}, schema={case_id_glue: case_id_dtype}).lazy()
    
    # Filter based on positive/negative
    if positive:
        # Keep only matching cases
        ret = df.join(matching_df, on=case_id_glue, how="inner")
    else:
        # Keep only non-matching cases
        ret = df.join(matching_df, on=case_id_glue, how="anti")
    
    return ret


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.objects.process_tree.importer.variants import ptml
from pm4py.util import exec_utils


class Variants(Enum):
    PTML = ptml


DEFAULT_VARIANT = Variants.PTML


def apply(file_path, variant=DEFAULT_VARIANT, parameters=None):
    """
    Imports a process tree from the specified path

    Parameters
    ---------------
    path
        Path
    variant
        Variant of the algorithm, possible values:
            - Variants.PTML
    parameters
        Possible parameters (version specific)

    Returns
    ---------------
    tree
        Process tree
    """
    return exec_utils.get_variant(variant).apply(
        file_path, parameters=parameters
    )


def deserialize(tree_string, variant=DEFAULT_VARIANT, parameters=None):
    """
    Deserialize a text/binary string representing a process tree in the PTML format

    Parameters
    ----------
    tree_string
        Process tree expressed as PTML string
    variant
        Variant of the algorithm, possible values:
            - Variants.PTML
    parameters
        Other parameters of the algorithm

    Returns
    ----------
    tree
        Process tree
    """
    return exec_utils.get_variant(variant).import_tree_from_string(
        tree_string, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

from pm4py.util import exec_utils, constants, xes_constants, pandas_utils


class Parameters(Enum):
    ARRIVAL_RATE = "arrival_rate"
    FINISH_RATE = "finish_rate"
    CASE_ID_COLUMN = constants.PARAMETER_CONSTANT_CASEID_KEY
    START_TIMESTAMP_COLUMN = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_COLUMN = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    RESOURCE_COLUMN = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    ACTIVITY_COLUMN = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    GROUPER_FREQ = "grouper_freq"
    SERVICE_TIME = "service_time"
    WAITING_TIME = "waiting_time"
    SOJOURN_TIME = "sojourn_time"
    DIFF_START_END = "diff_start_end"


_WINDOW_START_COL = "__window_start"


def _freq_to_polars_duration(freq: str) -> str:
    """
    Convert a pandas-style frequency string to a Polars duration string.
    """
    if not freq:
        raise ValueError("grouper frequency must be a non-empty string")

    freq = freq.strip()
    freq_upper = freq.upper()
    # Remove timezone/day designators (e.g., W-MON -> W)
    if "-" in freq_upper:
        freq_upper = freq_upper.split("-", 1)[0]

    numbers = "".join(ch for ch in freq_upper if ch.isdigit())
    letters = "".join(ch for ch in freq_upper if ch.isalpha())

    if not letters:
        raise ValueError(f"cannot infer unit from grouper frequency '{freq}'")

    if not numbers:
        numbers = "1"

    normalized_unit = letters
    if normalized_unit in {"MIN"}:
        normalized_unit = "MIN"
    else:
        # Drop suffixes such as S, E (month/quarter/year start/end)
        normalized_unit = normalized_unit.rstrip("S").rstrip("E")
        if not normalized_unit:
            normalized_unit = letters

    unit_mapping = {
        "S": "s",
        "SEC": "s",
        "SECOND": "s",
        "MIN": "m",
        "T": "m",
        "H": "h",
        "D": "d",
        "W": "w",
        "M": "mo",
        "Q": "q",
        "Y": "y",
        "A": "y",
    }

    if normalized_unit not in unit_mapping:
        raise ValueError(
            f"grouper frequency '{freq}' is not supported when using the temporal_lazy variant."
        )

    return f"{numbers}{unit_mapping[normalized_unit]}"


def apply(
    log,
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Extract temporal features using Polars lazy transformations.

    Parameters
    ----------------
    log
        Polars LazyFrame containing the event log
    parameters
        Parameters of the algorithm (same semantics as the pandas-based temporal variant)

    Returns
    ----------------
    pl.DataFrame
        Dataframe with temporal features, aggregated per time window
    """
    if parameters is None:
        parameters = {}

    import polars as pl

    if not pandas_utils.is_polars_lazyframe(log):
        raise TypeError(
            "temporal_lazy.apply expects a Polars LazyFrame as input."
        )

    grouper_freq = exec_utils.get_param_value(
        Parameters.GROUPER_FREQ, parameters, "W"
    )
    timestamp_column = exec_utils.get_param_value(
        Parameters.TIMESTAMP_COLUMN,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_column = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_COLUMN, parameters, None
    )
    if start_timestamp_column is None:
        start_timestamp_column = timestamp_column

    case_id_column = exec_utils.get_param_value(
        Parameters.CASE_ID_COLUMN, parameters, constants.CASE_CONCEPT_NAME
    )
    diff_start_end = exec_utils.get_param_value(
        Parameters.DIFF_START_END, parameters, "@@diff_start_end"
    )
    arrival_rate = exec_utils.get_param_value(
        Parameters.ARRIVAL_RATE, parameters, "@@arrival_rate"
    )
    finish_rate = exec_utils.get_param_value(
        Parameters.FINISH_RATE, parameters, "@@finish_rate"
    )
    service_time = exec_utils.get_param_value(
        Parameters.SERVICE_TIME, parameters, "@@service_time"
    )
    waiting_time = exec_utils.get_param_value(
        Parameters.WAITING_TIME, parameters, "@@waiting_time"
    )
    sojourn_time = exec_utils.get_param_value(
        Parameters.SOJOURN_TIME, parameters, "@@sojourn_time"
    )
    resource_column = exec_utils.get_param_value(
        Parameters.RESOURCE_COLUMN,
        parameters,
        xes_constants.DEFAULT_RESOURCE_KEY,
    )
    activity_column = exec_utils.get_param_value(
        Parameters.ACTIVITY_COLUMN, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    lazy_log = pandas_utils.insert_case_arrival_finish_rate(
        log,
        case_id_column=case_id_column,
        timestamp_column=timestamp_column,
        start_timestamp_column=start_timestamp_column,
        arrival_rate_column=arrival_rate,
        finish_rate_column=finish_rate,
    )

    lazy_log = pandas_utils.insert_case_service_waiting_time(
        lazy_log,
        case_id_column=case_id_column,
        timestamp_column=timestamp_column,
        start_timestamp_column=start_timestamp_column,
        diff_start_end_column=diff_start_end,
        service_time_column=service_time,
        sojourn_time_column=sojourn_time,
        waiting_time_column=waiting_time,
    )

    duration = _freq_to_polars_duration(grouper_freq)

    lazy_log = lazy_log.with_columns(
        pl.col(start_timestamp_column)
        .dt.truncate(duration)
        .alias(_WINDOW_START_COL)
    )

    case_agg = (
        lazy_log.group_by([_WINDOW_START_COL, case_id_column])
        .agg(
            pl.len().alias("__events_per_case"),
            pl.col(activity_column)
            .drop_nulls()
            .n_unique()
            .alias("__unique_acts_per_case"),
            pl.col(resource_column)
            .drop_nulls()
            .n_unique()
            .alias("__unique_res_per_case"),
            pl.col(arrival_rate).first().alias("__arrival_rate"),
            pl.col(finish_rate).first().alias("__finish_rate"),
            pl.col(waiting_time).first().alias("__waiting_time"),
            pl.col(sojourn_time).first().alias("__sojourn_time"),
            pl.col(service_time).first().alias("__service_time"),
        )
        .with_columns(
            (
                pl.col("__events_per_case")
                - pl.col("__unique_acts_per_case")
            ).alias("__reworks_per_case")
        )
    )

    case_window = case_agg.group_by(_WINDOW_START_COL).agg(
        pl.col("__reworks_per_case")
        .sum()
        .alias("total_number_of_reworked_activities"),
        pl.col("__events_per_case").mean().alias("avg_events_per_case"),
        pl.col("__unique_res_per_case").mean().alias("avg_resources_per_case"),
        pl.col(case_id_column).n_unique().alias("number_of_cases"),
        pl.col("__arrival_rate").mean().alias("average_arrival_rate"),
        pl.col("__finish_rate").mean().alias("average_finish_rate"),
        pl.col("__waiting_time").mean().alias("average_waiting_time"),
        pl.col("__sojourn_time").mean().alias("average_sojourn_time"),
        pl.col("__service_time").mean().alias("average_service_time"),
    )

    resource_agg = (
        lazy_log.filter(pl.col(resource_column).is_not_null())
        .group_by([_WINDOW_START_COL, resource_column])
        .agg(
            pl.col(case_id_column)
            .n_unique()
            .alias("__cases_per_resource")
        )
    )

    resource_window = resource_agg.group_by(_WINDOW_START_COL).agg(
        pl.col("__cases_per_resource")
        .mean()
        .alias("avg_cases_per_resource")
    )

    basic_window = lazy_log.group_by(_WINDOW_START_COL).agg(
        pl.col(resource_column)
        .drop_nulls()
        .n_unique()
        .alias("unique_resources"),
        pl.col(activity_column)
        .drop_nulls()
        .n_unique()
        .alias("unique_activities"),
        pl.len().alias("num_events"),
    )

    result_lazy = (
        basic_window.join(case_window, on=_WINDOW_START_COL, how="left")
        .join(resource_window, on=_WINDOW_START_COL, how="left")
        .sort(_WINDOW_START_COL)
        .rename({_WINDOW_START_COL: "timestamp"})
        .with_columns(
            [
                pl.col("total_number_of_reworked_activities").fill_null(0),
                pl.col("avg_events_per_case").fill_null(0),
                pl.col("avg_resources_per_case").fill_null(0),
                pl.col("average_arrival_rate").fill_null(0),
                pl.col("average_finish_rate").fill_null(0),
                pl.col("average_waiting_time").fill_null(0),
                pl.col("average_sojourn_time").fill_null(0),
                pl.col("average_service_time").fill_null(0),
                pl.col("avg_cases_per_resource").fill_null(0),
                pl.col("number_of_cases").fill_null(0),
                pl.col("unique_resources").fill_null(0),
                pl.col("unique_activities").fill_null(0),
                pl.col("num_events").fill_null(0),
            ]
        )
        .select(
            [
                "timestamp",
                "unique_resources",
                "unique_activities",
                "num_events",
                "average_arrival_rate",
                "average_finish_rate",
                "average_waiting_time",
                "average_sojourn_time",
                "average_service_time",
                "total_number_of_reworked_activities",
                "avg_cases_per_resource",
                "avg_events_per_case",
                "number_of_cases",
                "avg_resources_per_case",
            ]
        )
    )

    return result_lazy



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Tuple, List, Any, Set, Optional, Dict, Union

from pm4py.util import exec_utils
import heapq
from copy import copy


class Parameters(Enum):
    MERGE_DISTANCE = "merge_distance"
    MIN_BATCH_SIZE = "min_batch_size"


class BatchType(Enum):
    SIMULTANEOUS = "Simultaneous"
    BATCHING_START = "Batching on Start"
    BATCHING_END = "Batching on End"
    SEQ_BATCHING = "Sequential batching"
    CONC_BATCHING = "Concurrent batching"


def __merge_overlapping_intervals(
    intervals: List[Tuple[float, float, Set[Any]]]
) -> List[Tuple[float, float, Set[Any]]]:
    """
    Iterative method that merges the overlapping time intervals
    (an interval [a, b] is overlapping to [c, d] if a <= c <= b or c <= a <= d).
    """
    continue_cycle = True
    while continue_cycle:
        continue_cycle = False
        i = 0
        while i < len(intervals) - 1:
            if intervals[i][1] > intervals[i + 1][0]:
                # decide to merge interval i and i+1
                new_interval = (
                    min(intervals[i][0], intervals[i + 1][0]),
                    max(intervals[i][1], intervals[i + 1][1]),
                    intervals[i][2].union(intervals[i + 1][2]),
                )
                # add the new interval to the list
                intervals.append(new_interval)
                # remove the i+1 interval
                del intervals[i + 1]
                # remove the i interval
                del intervals[i]
                # sort the intervals
                intervals.sort()
                # set the variable continue_cycle to True
                continue_cycle = True
                # interrupt the current iteration on the intervals
                break
            i = i + 1
    return intervals


def __merge_near_intervals(
    intervals: List[Tuple[float, float, Set[Any]]], max_allowed_distance: float
) -> List[Tuple[float, float, Set[Any]]]:
    """
    Merge the non-overlapping time intervals that are nearer than max_allowed_distance.
    (an interval [a, b] that is non-overlapping with [c, d] having b < c, is merged if c - b <= max_allowed_distance).
    """
    continue_cycle = True
    while continue_cycle:
        continue_cycle = False
        i = 0
        while i < len(intervals) - 1:
            if intervals[i + 1][0] - intervals[i][1] <= max_allowed_distance:
                # decide to merge interval i and i+1
                new_interval = (
                    min(intervals[i][0], intervals[i + 1][0]),
                    max(intervals[i][1], intervals[i + 1][1]),
                    intervals[i][2].union(intervals[i + 1][2]),
                )
                # remove the i+1 interval
                del intervals[i + 1]
                # remove the i interval
                del intervals[i]
                # add the new interval to the list
                heapq.heappush(intervals, new_interval)
                # set the variable continue_cycle to True
                continue_cycle = True
                i = i - 1
            i = i + 1
    return intervals


def __check_batch_type(batch: Tuple[float, float, Set[Any]]) -> str:
    """
    Checks the batch type between:
    - Simultaneous (all the events in the batch have identical start and end timestamps)
    - Batching at start (all the events in the batch have identical start timestamp)
    - Batching at end (all the events in the batch have identical end timestamp)
    - Sequential batching (for all the consecutive events, the end of the first is equal to the start of the second)
    - Concurrent batching (for all the consecutive events that are not sequentially matched)
    """
    events_batch = sorted(list(batch[2]))
    # take the minimum of the left-extreme of each interval
    min_left_events = min(ev[0] for ev in events_batch)
    # take the maximum of the left-extreme of each interval
    max_left_events = max(ev[0] for ev in events_batch)
    # take the minimum of the right-extreme of each interval
    min_right_events = min(ev[1] for ev in events_batch)
    # take the maximum of the right-extreme of each interval
    max_right_events = max(ev[1] for ev in events_batch)

    # CONDITION 1 - All the events in the batch have identical start and end
    # timestamps
    if (
        min_left_events == max_left_events
        and min_right_events == max_right_events
    ):
        return BatchType.SIMULTANEOUS.value
    # CONDITION 4 - All the events in the batch have identical start timestamp:
    if min_left_events == max_left_events:
        return BatchType.BATCHING_START.value
    # CONDITION 5 - All the events in the batch have identical end timestamp:
    if min_right_events == max_right_events:
        return BatchType.BATCHING_END.value

    # now we could be in the SEQUENTIAL batching or the CONCURRENT batching
    # in order to be in the SEQUENTIAL, we need that for all the consecutive
    # events the end of the first is equal to the start of the second
    is_sequential = True
    i = 0
    while i < len(events_batch) - 1:
        # if there are two consecutive events that are not sequentially
        # matched, then we automatically fall inside the CONCURRENT batching
        if events_batch[i][1] != events_batch[i + 1][0]:
            is_sequential = False
            break
        i = i + 1
    if is_sequential:
        return BatchType.SEQ_BATCHING.value
    else:
        return BatchType.CONC_BATCHING.value


def __detect_single(
    events: List[Tuple[float, float, str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, List[Any]]:
    """
    Detect if there are batches in the execution of events having a given activity-resource combination
    """
    if parameters is None:
        parameters = {}

    ret = {
        BatchType.SIMULTANEOUS.value: [],
        BatchType.BATCHING_START.value: [],
        BatchType.BATCHING_END.value: [],
        BatchType.CONC_BATCHING.value: [],
        BatchType.SEQ_BATCHING.value: [],
    }
    merge_distance = exec_utils.get_param_value(
        Parameters.MERGE_DISTANCE, parameters, 15 * 60
    )
    min_batch_size = exec_utils.get_param_value(
        Parameters.MIN_BATCH_SIZE, parameters, 2
    )

    intervals = [(e[0], e[1], {copy(e)}) for e in events]
    heapq.heapify(intervals)
    intervals = __merge_overlapping_intervals(intervals)
    intervals = __merge_near_intervals(intervals, merge_distance)
    batches = [x for x in intervals if len(x[2]) >= min_batch_size]
    for batch in batches:
        batch_type = __check_batch_type(batch)
        ret[batch_type].append(batch)
    ret = {x: y for x, y in ret.items() if y}
    return ret


def detect(
    actres_grouping: Dict[Tuple[str, str], List[Tuple[float, float, str]]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Tuple[Tuple[str, str], int, Dict[str, Any]]]:
    """
    Provided an activity-resource grouping of the events of the event log, returns
    a list having as elements the activity-resources with the batches that are detected, divided in:
    - Simultaneous (all the events in the batch have identical start and end timestamps)
    - Batching at start (all the events in the batch have identical start timestamp)
    - Batching at end (all the events in the batch have identical end timestamp)
    - Sequential batching (for all the consecutive events, the end of the first is equal to the start of the second)
    - Concurrent batching (for all the consecutive events that are not sequentially matched)

    The approach has been described in the following paper:
    Martin, N., Swennen, M., Depaire, B., Jans, M., Caris, A., & Vanhoof, K. (2015, December). Batch Processing:
    Definition and Event Log Identification. In SIMPDA (pp. 137-140).

    Parameters
    -------------------
    actres_grouping
        Activity-resource grouping of events
    parameters
        Parameters of the algorithm

    Returns
    ------------------
    list_batches
        A (sorted) list containing tuples. Each tuple contain:
        - Index 0: the activity-resource for which at least one batch has been detected
        - Index 1: the number of batches for the given activity-resource
        - Index 2: a list containing all the batches. Each batch is described by:
            # The start timestamp of the batch
            # The complete timestamp of the batch
            # The list of events that are executed in the batch
    """
    if parameters is None:
        parameters = {}
    ret = []
    for actres in actres_grouping:
        batches = __detect_single(
            actres_grouping[actres], parameters=parameters
        )
        if batches:
            total_length = sum(len(y) for y in batches.values())
            ret.append((actres, total_length, batches))
    ret = sorted(ret, reverse=True, key=lambda x: (x[1], x[0]))
    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import copy
import itertools
import random
from abc import ABC, abstractclassmethod
from collections import Counter
from typing import Counter as TCounter
from typing import Generic, List, Optional, Tuple, TypeVar

from pm4py.objects.petri_net.saw_net.obj import StochasticArcWeightNet
from pm4py.objects.petri_net.stochastic.semantics import (
    StochasticPetriNetSemantics,
)

N = TypeVar("N", bound=StochasticArcWeightNet)
T = TypeVar("T", bound=StochasticArcWeightNet.Transition)
P = TypeVar("P", bound=StochasticArcWeightNet.Place)
A = TypeVar("A", bound=StochasticArcWeightNet.Arc)
B = TypeVar("B", bound=StochasticArcWeightNet.Binding)


class StochasticArcWeightNetSemantics(
    StochasticPetriNetSemantics[N], Generic[N], ABC
):

    @classmethod
    def is_enabled(cls, pn: N, transition: T, marking: TCounter[P]) -> bool:
        """
        Checks whether a given transition is enabled in a given Petri net and marking.
        Every place should at least have the same number of tokens as the  minimum binding that has a weight above 0

        Parameters
        ----------
        :param pn: Petri net
        :param transition: transition to check
        :param marking: marking to check

        Returns
        -------
        :return: true if enabled, false otherwise
        """
        if transition not in pn.transitions:
            return False
        for a in transition.in_arcs:
            if marking[a.source] < min(
                [k for k, v in a.weight_distribution.items() if v > 0]
            ):
                return False
        return True

    @classmethod
    def fire(cls, pn: N, binding: B, marking: TCounter[P]) -> TCounter[P]:
        """fires the binding in the given marking. Does not check if the the binding is feasible (this should be handled by the invoking code)

        Args:
            pn (N): saw net to use
            marking (TCounter[P]): marking to use
            binding (B): binding to use

        Returns:
            TCounter[P]: _description_
        """
        m_out = copy.copy(marking)
        for a, w in binding:
            if isinstance(a.source, StochasticArcWeightNet.Place):
                m_out[a.source] -= w
            else:
                m_out[a.target] += w
        return m_out

    @classmethod
    def all_legal_bindings(
        cls, pn: N, transition: T
    ) -> List[List[Tuple[A, int]]]:
        """
        Creates all possible bindings for a given input transition

        Parameters
        ----------
        :param pn: Petri net
        :param transition: transition to genereate all bindings for

        Returns
        -------
        :return: list containing all posible bindings
        """
        if transition not in pn.transitions:
            return []
        else:
            l = list()
            for a in transition.in_arcs:
                l.append(
                    [(a, k) for k, v in a.weight_distribution.items() if v > 0]
                )
            for a in transition.out_arcs:
                l.append(
                    [(a, k) for k, v in a.weight_distribution.items() if v > 0]
                )
            return list(itertools.product(*l))

    @classmethod
    def all_enabled_bindings(
        cls, pn: N, transition: T, marking: TCounter[P]
    ) -> List[List[Tuple[A, int]]]:
        """
        Creates all possible feasible bindings for a given input transition in a given marking

        Parameters
        ----------
        :param pn: Petri net
        :param marking: marking to use
        :param transition: transition to genereate all feasible bindings for

        Returns
        -------
        :return: list containing all posible feasible bindings
        """
        return list(
            filter(
                lambda b: cls.is_enabled_binding(pn, transition, b, marking),
                cls.all_legal_bindings(pn, transition),
            )
        )

    @classmethod
    def is_enabled_binding(
        cls,
        pn: N,
        transition: T,
        binding: StochasticArcWeightNet.Binding,
        marking: TCounter[P],
    ) -> bool:
        """
        Checks if the provided binding is enabled

        Parameters
        ----------
        :param pn: Petri net
        :param marking: marking to use
        :param transition: transition to genereate all feasible bindings for

        Returns
        -------
        :return: bool indicates if the binding is enabled
        """
        if transition not in pn.transitions:
            return False
        places_in_bindings = set(
            filter(
                lambda x: isinstance(x, StochasticArcWeightNet.Place),
                [
                    x
                    for x in list(
                        itertools.chain(
                            *[(a.source, a.target) for (a, w) in binding]
                        )
                    )
                ],
            )
        )
        for a in transition.in_arcs:
            if a.source not in places_in_bindings:
                return False
        for a in transition.out_arcs:
            if a.target not in places_in_bindings:
                return False
        for a, w in binding:
            if a.weight_distribution[w] == 0.0:
                return False
            if transition not in {a.source, a.target}:
                return False
            if transition == a.target:
                if w > marking[a.source]:
                    return False
        return True

    @classmethod
    def amortized_priority(
        cls, binding: StochasticArcWeightNet.Binding
    ) -> float:
        """
        Computes the amortized priority (a.k.a weight) of a binding. The amortized priority is equal to the product of all individual weights of the arc weights includec in the binding.

        Args:
            binding (StochasticArcWeightNet.Binding): input binding

        Returns:
            float: amortized weight
        """
        prod = 1
        for a, w in binding:
            prod *= a.weight_distribution[w]
        return prod

    @abstractclassmethod
    def probability_of_binding(
        cls,
        pn: N,
        transition: T,
        binding: StochasticArcWeightNet.Binding,
        marking: TCounter[P],
    ) -> float:
        """
        Calculates the probability of firing a transition t under binding b in the net, in the given marking.

        Parameters
        ----------
        :param pn: Petri net
        :param transition: transition to fire
        :param binding: binding to consider
        :param marking: marking to use

        Returns
        -------
        :return: firing probability of transition t under binding b
        """
        pass


class LocalStochasticArcWeightNetSemantics(
    StochasticArcWeightNetSemantics[N], Generic[N]
):

    @classmethod
    def probability_of_binding(
        cls,
        pn: N,
        transition: T,
        binding: StochasticArcWeightNet.Binding,
        marking: TCounter[P],
    ) -> float:
        """
        Calculates the probability of firing a transition t under binding b in the net, in the given marking.

        Parameters
        ----------
        :param pn: Petri net
        :param transition: transition to fire
        :param binding: binding to consider
        :param marking: marking to use

        Returns
        -------
        :return: firing probability of transition t under binding b
        """
        s = 0
        for t in pn.transitions:
            if cls.is_enabled(pn, t, marking):
                for b in cls.all_enabled_bindings(pn, transition, marking):
                    s += t.weight * cls.amortized_priority(b)

        print(cls.amortized_priority(binding))
        return transition.weight * cls.amortized_priority(binding) / s


class GlobalStochasticArcWeightNetSemantics(
    StochasticArcWeightNetSemantics[N], Generic[N]
):

    @classmethod
    def probability_of_transition(
        cls, pn: N, transition: T, marking: TCounter[P]
    ) -> float:
        """
        Compute the probability of firing a transition in the net and marking.

        Args:
            pn (N): Stochastic net
            transition (T): transition to fire
            marking (Counter[P]): marking to use

        Returns:
            float: _description_
        """
        return (
            0.0
            if transition not in pn.transitions
            or not cls.is_enabled(pn, transition, marking)
            else sum(
                [
                    transition.weight * cls.amortized_priority(b)
                    for b in cls.all_enabled_bindings(pn, transition, marking)
                ]
            )
            / sum(
                [
                    t.weight * cls.amortized_priority(b)
                    for t in pn.transitions
                    if cls.is_enabled(pn, t, marking)
                    for b in cls.all_enabled_bindings(pn, t, marking)
                ]
            )
        )

    @classmethod
    def sample_enabled_transition(
        cls, pn: N, marking: TCounter[P], seed: int = None
    ) -> Optional[Tuple[T, B]]:
        if seed is not None:
            random.seed(seed)
        bindings = list()
        probs = list()
        for t in [t for t in pn.transitions if cls.is_enabled(pn, t, marking)]:
            for b in cls.all_enabled_bindings(pn, t, marking):
                bindings.append((t, b))
                probs.append(cls.probability_of_binding(pn, t, b, marking))
        return (
            None if len(bindings) == 0 else random.choices(bindings, probs)[0]
        )

    @classmethod
    def probability_of_binding(
        cls,
        pn: N,
        transition: T,
        binding: StochasticArcWeightNet.Binding,
        marking: TCounter[P],
    ) -> float:
        """
        Calculates the probability of firing a transition t under binding b in the net, in the given marking.

        Parameters
        ----------
        :param pn: Petri net
        :param transition: transition to fire
        :param binding: binding to consider
        :param marking: marking to use

        Returns
        -------
        :return: firing probability of transition t under binding b
        """
        return (
            0.0
            if transition not in pn.transitions
            or not cls.is_enabled_binding(pn, transition, binding, marking)
            else (transition.weight * cls.amortized_priority(binding))
            / sum(
                [
                    t.weight * cls.amortized_priority(b)
                    for t in pn.transitions
                    if cls.is_enabled(pn, t, marking)
                    for b in cls.all_enabled_bindings(pn, t, marking)
                ]
            )
        )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.petri_net.utils.performance_map import (
    calculate_annotation_for_trace,
    single_element_statistics,
    find_min_max_trans_frequency,
    find_min_max_arc_frequency,
    aggregate_stats,
    find_min_max_arc_performance,
    aggregate_statistics,
    get_transition_performance_with_token_replay,
    get_idx_exceeding_specified_acti_performance,
    filter_cases_exceeding_specified_acti_performance,
)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.algo.conformance.footprints.variants import (
    log_model,
    log_extensive,
    trace_extensive,
)
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List


class Variants(Enum):
    LOG_MODEL = log_model
    LOG_EXTENSIVE = log_extensive
    TRACE_EXTENSIVE = trace_extensive


def apply(
    log_footprints: Union[Dict[str, Any], List[Dict[str, Any]]],
    model_footprints: Dict[str, Any],
    variant=Variants.LOG_MODEL,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    """
    Apply footprints conformance between a log footprints object
    and a model footprints object

    Parameters
    -----------------
    log_footprints
        Footprints of the log
    model_footprints
        Footprints of the model
    parameters
        Parameters of the algorithm, including:
            - Parameters.STRICT => strict check of the footprints

    Returns
    ------------------
    violations
        Set/dictionary of all the violations between the log footprints
        and the model footprints, OR list of case-per-case violations
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(
        log_footprints, model_footprints, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.streaming.algo.conformance.declare.variants import automata


class Variants(Enum):
    AUTOMATA = automata


def apply(declare_model, variant=Variants.AUTOMATA, parameters=None):
    """
    Streaming Conformance Checking Algorithm for DECLARE models.
    Attempts to implement state-based checks for all Declare constraint types.
    When a violation occurs, prints out which constraints are violated.

    Streaming algorithm interface implemented.

    Implementation of:
    Maggi, Fabrizio Maria, et al. "Monitoring business constraints with linear temporal logic: An approach based on colored automata." Business Process Management: 9th International Conference, BPM 2011, Clermont-Ferrand, France, August 30-September 2, 2011. Proceedings 9. Springer Berlin Heidelberg, 2011.
    -------

    """
    return exec_utils.get_variant(variant).apply(declare_model, parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.obj import EventLog, Trace, Event
from copy import copy


def keep_one_trace_per_variant(log, parameters=None):
    """
    Keeps only one trace per variant (does not matter for basic inductive miner)

    Parameters
    --------------
    log
        Log
    parameters
        Parameters of the algorithm

    Returns
    --------------
    new_log
        Log (with one trace per variant)
    """
    if parameters is None:
        parameters = {}

    from pm4py.statistics.variants.log import get as variants_module

    new_log = EventLog()
    if log is not None:
        variants = variants_module.get_variants(log, parameters=parameters)
        for var in variants:
            curr_trace = variants[var][0]
            new_trace = Trace(attributes=copy(curr_trace.attributes))
            new_trace.attributes["@@num_traces"] = len(variants[var])
            for ev in curr_trace:
                new_trace.append(ev)
            new_log.append(new_trace)

    return new_log


def keep_only_one_attribute_per_event(log, attribute_key):
    """
    Keeps only one attribute per event

    Parameters
    ---------------
    log
        Event log
    attribute_key
        Attribute key
    """
    new_log = EventLog()
    if log is not None:
        for trace in log:
            new_trace = Trace()
            for ev in trace:
                new_trace.append(Event({attribute_key: ev[attribute_key]}))
            new_log.append(new_trace)

    return new_log



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
import numpy as np
from enum import Enum
from typing import Optional, Dict, Any
from pm4py.util import exec_utils


class Parameters(Enum):
    MAX_DIVISIONS_X = "max_divisions_x"
    MAX_DIVISIONS_Y = "max_divisions_y"
    AGGREGATION_FUNCTION = "aggregation_function"
    X_BINS = "x_bins"           # Optional list of numeric bin edges for x_col
    Y_BINS = "y_bins"           # Optional list of numeric bin edges for y_col


def apply(
        feature_table: pd.DataFrame,
        x_col: str,
        y_col: str,
        agg_col: str,
        parameters: Optional[Dict[Any, Any]] = None
):
    """
    Constructs a process cube by slicing data along two dimensions
    (x_col, y_col) and aggregating a third (agg_col). Additionally:

    1) If x_col (or y_col) is an actual column in df, we do numeric binning.
       You can manually specify bin edges via parameters[Parameters.X_BINS]
       (a list of numeric edges) or parameters[Parameters.Y_BINS].
       Otherwise, we automatically divide into equal-width bins using
       parameters[Parameters.MAX_DIVISIONS_X] or MAX_DIVISIONS_Y.
    2) If x_col (or y_col) is not present, we do prefix-based binning.

    Parameters
    ----------
    feature_table : pd.DataFrame
        A feature table that must contain 'case:concept:name' and agg_col, plus
        the columns for x_col, y_col (if numeric) or the columns that start
        with x_col, y_col (if prefix-based).
    x_col : str
        The X dimension. If x_col in df.columns, numeric binning; else prefix-based.
    y_col : str
        The Y dimension. If y_col in df.columns, numeric binning; else prefix-based.
    agg_col : str
        The column to aggregate (mean, sum, etc.).
    parameters: Dict[Any, Any]
        Optional parameters of the method, including:
        * Parameters.X_BINS: List of numeric bin edges for x_col.
        * Parameters.Y_BINS: List of numeric bin edges for y_col.
        * Parameters.MAX_DIVISIONS_X: If x_col is numeric and X_BINS not provided,
          how many bins to divide it into.
        * Parameters.MAX_DIVISIONS_Y: If y_col is numeric and Y_BINS not provided,
          how many bins to divide it into.
        * Parameters.AGGREGATION_FUNCTION: The aggregation function,
          e.g., 'mean', 'sum', 'min', 'max'.

    Returns
    -------
    pivot_df : pd.DataFrame
        A pivoted DataFrame representing the process cube, with x bins as rows
        and y bins as columns, containing aggregated values of agg_col.
    cell_case_dict : dict
        A dictionary mapping (x_bin, y_bin) -> set of case IDs that fall in that cell.
    """
    if parameters is None:
        parameters = {}

    # Retrieve parameters, with None defaults for manual bins
    max_divisions_x = exec_utils.get_param_value(Parameters.MAX_DIVISIONS_X, parameters, 4)
    max_divisions_y = exec_utils.get_param_value(Parameters.MAX_DIVISIONS_Y, parameters, 4)
    agg_fn = exec_utils.get_param_value(Parameters.AGGREGATION_FUNCTION, parameters, "mean")
    x_bins_param = exec_utils.get_param_value(Parameters.X_BINS, parameters, None)
    y_bins_param = exec_utils.get_param_value(Parameters.Y_BINS, parameters, None)

    # Work with a view instead of copy when possible
    df = feature_table
    
    # Pre-compute column lists and masks for better performance
    numeric_x = x_col in df.columns
    numeric_y = y_col in df.columns

    x_prefix_cols = []
    y_prefix_cols = []
    if not numeric_x:
        x_prefix_cols = [c for c in df.columns if c.startswith(x_col+"_")]
    if not numeric_y:
        y_prefix_cols = [c for c in df.columns if c.startswith(y_col+"_")]

    x_all_bins = None
    y_all_bins = None

    # ------------------------------------------------------
    # Handle X dimension binning
    # ------------------------------------------------------
    if numeric_x:
        # Use manual bins if provided, else auto-generate equal-width bins
        if x_bins_param is not None:
            x_bins = sorted(list(set(x_bins_param)))  # Remove duplicates and sort
        else:
            x_min, x_max = df[x_col].min(), df[x_col].max()
            if x_min == x_max:
                # Handle case where all values are the same
                x_bins = [x_min - 0.5, x_max + 0.5]
            else:
                x_bins = np.linspace(x_min, x_max, max_divisions_x + 1)
                # Ensure bins are unique
                x_bins = np.unique(x_bins)
        
        # Create binned column directly without temporary column
        x_binned = pd.cut(df[x_col], bins=x_bins, include_lowest=True)
        x_valid_mask = pd.notna(x_binned)
        x_all_bins = x_binned.cat.categories
    else:
        # Pre-filter and vectorize prefix-based column selection
        x_prefix_data = df[x_prefix_cols].fillna(0)
        x_valid_cols_mask = x_prefix_data >= 1
        x_valid_mask = x_valid_cols_mask.any(axis=1)
        x_all_bins = x_prefix_cols

    # ------------------------------------------------------
    # Handle Y dimension binning
    # ------------------------------------------------------
    if numeric_y:
        if y_bins_param is not None:
            y_bins = sorted(list(set(y_bins_param)))  # Remove duplicates and sort
        else:
            y_min, y_max = df[y_col].min(), df[y_col].max()
            if y_min == y_max:
                # Handle case where all values are the same
                y_bins = [y_min - 0.5, y_max + 0.5]
            else:
                y_bins = np.linspace(y_min, y_max, max_divisions_y + 1)
                # Ensure bins are unique
                y_bins = np.unique(y_bins)
        
        y_binned = pd.cut(df[y_col], bins=y_bins, include_lowest=True)
        y_valid_mask = pd.notna(y_binned)
        y_all_bins = y_binned.cat.categories
    else:
        y_prefix_data = df[y_prefix_cols].fillna(0)
        y_valid_cols_mask = y_prefix_data >= 1
        y_valid_mask = y_valid_cols_mask.any(axis=1)
        y_all_bins = y_prefix_cols

    # Combined validity mask
    valid_mask = x_valid_mask & y_valid_mask
    if not valid_mask.any():
        return pd.DataFrame(), {}

    # Filter data to valid rows only
    valid_df = df[valid_mask]
    case_ids = valid_df["case:concept:name"].values
    agg_values = valid_df[agg_col].values

    # Build DataFrame directly using vectorized operations - much faster than building records list
    if numeric_x and numeric_y:
        # Both numeric - create DataFrame directly
        x_bins_valid = x_binned[valid_mask]
        y_bins_valid = y_binned[valid_mask]
        
        temp_df = pd.DataFrame({
            "case:concept:name": case_ids,
            "x_bin": x_bins_valid,
            "y_bin": y_bins_valid,
            agg_col: agg_values
        })
        
    elif numeric_x and not numeric_y:
        # X numeric, Y prefix-based - use more efficient vectorized approach
        x_bins_valid = x_binned[valid_mask]
        y_valid_cols_valid = y_valid_cols_mask[valid_mask]
        
        # Use numpy operations for much faster processing
        y_valid_array = y_valid_cols_valid.values
        row_counts = np.sum(y_valid_array, axis=1)
        total_rows = np.sum(row_counts)
        
        if total_rows == 0:
            return pd.DataFrame(), {}
        
        # Pre-allocate arrays for better performance
        case_ids_expanded = np.empty(total_rows, dtype=object)
        x_bins_expanded = np.empty(total_rows, dtype=object)
        y_cols_expanded = np.empty(total_rows, dtype=object)
        agg_values_expanded = np.empty(total_rows, dtype=float)
        
        # Fill arrays using vectorized operations
        idx = 0
        y_prefix_cols_array = np.array(y_prefix_cols)
        
        for i in range(len(case_ids)):
            if row_counts[i] > 0:
                valid_y_indices = np.where(y_valid_array[i])[0]
                n_valid = len(valid_y_indices)
                
                case_ids_expanded[idx:idx+n_valid] = case_ids[i]
                x_bins_expanded[idx:idx+n_valid] = x_bins_valid.iloc[i]
                y_cols_expanded[idx:idx+n_valid] = y_prefix_cols_array[valid_y_indices]
                agg_values_expanded[idx:idx+n_valid] = agg_values[i]
                idx += n_valid
            
        temp_df = pd.DataFrame({
            "case:concept:name": case_ids_expanded,
            "x_bin": x_bins_expanded,
            "y_bin": y_cols_expanded,
            agg_col: agg_values_expanded
        })
        
    elif not numeric_x and numeric_y:
        # X prefix-based, Y numeric - use more efficient vectorized approach
        x_valid_cols_valid = x_valid_cols_mask[valid_mask]
        y_bins_valid = y_binned[valid_mask]
        
        # Use numpy operations for much faster processing
        x_valid_array = x_valid_cols_valid.values
        row_counts = np.sum(x_valid_array, axis=1)
        total_rows = np.sum(row_counts)
        
        if total_rows == 0:
            return pd.DataFrame(), {}
        
        # Pre-allocate arrays for better performance
        case_ids_expanded = np.empty(total_rows, dtype=object)
        x_cols_expanded = np.empty(total_rows, dtype=object)
        y_bins_expanded = np.empty(total_rows, dtype=object)
        agg_values_expanded = np.empty(total_rows, dtype=float)
        
        # Fill arrays using vectorized operations
        idx = 0
        x_prefix_cols_array = np.array(x_prefix_cols)
        
        for i in range(len(case_ids)):
            if row_counts[i] > 0:
                valid_x_indices = np.where(x_valid_array[i])[0]
                n_valid = len(valid_x_indices)
                
                case_ids_expanded[idx:idx+n_valid] = case_ids[i]
                x_cols_expanded[idx:idx+n_valid] = x_prefix_cols_array[valid_x_indices]
                y_bins_expanded[idx:idx+n_valid] = y_bins_valid.iloc[i]
                agg_values_expanded[idx:idx+n_valid] = agg_values[i]
                idx += n_valid
            
        temp_df = pd.DataFrame({
            "case:concept:name": case_ids_expanded,
            "x_bin": x_cols_expanded,
            "y_bin": y_bins_expanded,
            agg_col: agg_values_expanded
        })
        
    else:
        # Both prefix-based - most complex case, use highly optimized vectorized approach
        x_valid_cols_valid = x_valid_cols_mask[valid_mask]
        y_valid_cols_valid = y_valid_cols_mask[valid_mask]
        
        # Use numpy operations for much faster processing
        x_valid_array = x_valid_cols_valid.values
        y_valid_array = y_valid_cols_valid.values
        
        # Calculate total number of combinations for pre-allocation
        x_row_counts = np.sum(x_valid_array, axis=1)
        y_row_counts = np.sum(y_valid_array, axis=1)
        row_combinations = x_row_counts * y_row_counts
        total_rows = np.sum(row_combinations)
        
        if total_rows == 0:
            return pd.DataFrame(), {}
        
        # Pre-allocate arrays for maximum performance
        case_ids_expanded = np.empty(total_rows, dtype=object)
        x_cols_expanded = np.empty(total_rows, dtype=object)
        y_cols_expanded = np.empty(total_rows, dtype=object)
        agg_values_expanded = np.empty(total_rows, dtype=float)
        
        # Use vectorized operations with pre-converted arrays
        x_prefix_cols_array = np.array(x_prefix_cols)
        y_prefix_cols_array = np.array(y_prefix_cols)
        
        idx = 0
        for i in range(len(case_ids)):
            if row_combinations[i] > 0:
                valid_x_indices = np.where(x_valid_array[i])[0]
                valid_y_indices = np.where(y_valid_array[i])[0]
                
                # Create cartesian product using numpy operations
                x_mesh, y_mesh = np.meshgrid(valid_x_indices, valid_y_indices, indexing='ij')
                x_flat = x_mesh.flatten()
                y_flat = y_mesh.flatten()
                n_combinations = len(x_flat)
                
                # Fill arrays efficiently
                case_ids_expanded[idx:idx+n_combinations] = case_ids[i]
                x_cols_expanded[idx:idx+n_combinations] = x_prefix_cols_array[x_flat]
                y_cols_expanded[idx:idx+n_combinations] = y_prefix_cols_array[y_flat]
                agg_values_expanded[idx:idx+n_combinations] = agg_values[i]
                idx += n_combinations
            
        temp_df = pd.DataFrame({
            "case:concept:name": case_ids_expanded,
            "x_bin": x_cols_expanded,
            "y_bin": y_cols_expanded,
            agg_col: agg_values_expanded
        })

    if temp_df.empty:
        return pd.DataFrame(), {}

    # Optimized aggregation using more efficient groupby operations
    grouped = temp_df.groupby(["x_bin", "y_bin"], sort=False)
    
    # Compute aggregations separately for better performance
    agg_values_result = grouped[agg_col].agg(agg_fn).reset_index()
    case_sets_result = grouped["case:concept:name"].apply(lambda x: set(x)).reset_index()
    case_sets_result.rename(columns={"case:concept:name": "case_set"}, inplace=True)
    
    # Merge results efficiently
    agg_result = pd.merge(agg_values_result, case_sets_result, on=["x_bin", "y_bin"])

    # Use pivot_table directly for better performance and handling of missing values
    pivot_df = temp_df.pivot_table(
        index="x_bin", 
        columns="y_bin", 
        values=agg_col, 
        aggfunc=agg_fn,
        dropna=False
    )

    # Always keep all configured bins, even when no entries end up in that slice
    if x_all_bins is not None and len(x_all_bins) > 0:
        pivot_df = pivot_df.reindex(x_all_bins)
    if y_all_bins is not None and len(y_all_bins) > 0:
        pivot_df = pivot_df.reindex(columns=y_all_bins)

    # Build cell-case mapping using vectorized operations
    valid_x_mask = agg_result["x_bin"].isin(pivot_df.index)
    valid_y_mask = agg_result["y_bin"].isin(pivot_df.columns)
    valid_mask = valid_x_mask & valid_y_mask
    
    valid_agg_result = agg_result[valid_mask]
    cell_case_dict = dict(zip(
        zip(valid_agg_result["x_bin"], valid_agg_result["y_bin"]),
        valid_agg_result["case_set"]
    ))

    return pivot_df, cell_case_dict



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from typing import Optional, Tuple, List, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough
from pm4py.objects.process_tree.obj import ProcessTree, Operator
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL


class StrictTauLoopUVCL(FallThrough[IMDataStructureUVCL]):

    @classmethod
    def _get_projected_log(
        cls, log: UVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> UVCL:
        start_activities = comut.get_start_activities(log)
        end_activities = comut.get_end_activities(log)
        proj = Counter()
        for t in log:
            x = 0
            for i in range(1, len(t)):
                if t[i] in start_activities and t[i - 1] in end_activities:
                    proj.update({t[x:i]: log[t]})
                    x = i
            proj.update({t[x: len(t)]: log[t]})
        return proj

    @classmethod
    def holds(
        cls,
        obj: IMDataStructureUVCL,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> bool:
        log = obj.data_structure
        return sum(cls._get_projected_log(log).values()) > sum(log.values())

    @classmethod
    def apply(
        cls,
        obj: IMDataStructureUVCL,
        pool=None,
        manager=None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[IMDataStructureUVCL]]]:
        log = obj.data_structure
        proj = cls._get_projected_log(log)
        if sum(proj.values()) > sum(log.values()):
            return ProcessTree(operator=Operator.LOOP), [
                IMDataStructureUVCL(proj),
                IMDataStructureUVCL(Counter()),
            ]



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, constants
from typing import Optional, Dict, Any
import graphviz
import tempfile
import uuid
import networkx as nx


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    INCLUDE_NODE_ATTRIBUTES = "include_node_attributes"
    INCLUDE_EDGE_ATTRIBUTES = "include_edge_attributes"


def apply(
    G: nx.DiGraph, parameters: Optional[Dict[Any, Any]] = None
) -> graphviz.Digraph:
    """
    Creates a Graphviz Digraph from a NetworkX DiGraph object.

    Minimum viable example:

        import pm4py
        from pm4py.visualization.networkx import visualizer as nx_to_gv_vis

        log = pm4py.read_xes("../tests/input_data/running-example.xes")
        # gets an 'event graph' where events, cases and their relationships
        # are represented in a graph (NetworkX DiGraph)
        event_graph = pm4py.convert_log_to_networkx(log)

        # visualize the NX DiGraph using Graphviz
        gviz = nx_to_gv_vis.apply(event_graph, parameters={"format": "svg"})
        nx_to_gv_vis.view(gviz)


    Parameters
    ---------------
    G
        NetworkX DiGraph
    parameters
        Parameters of the visualization, including:
        - Parameters.FORMAT => format of the visualization (.png, .svg)
        - Parameters.BGCOLOR => background color of the visualization (transparent, white)
        - Parameters.RANKDIR => direction of the graph (LR or TB)
        - Parameters.INCLUDE_NODE_ATTRIBUTES => includes the node attributes in the node's label in the representation
        - Parameters.INCLUDE_EDGE_ATTRIBUTES => includes the edge attributes in the edge's label in the representation

    Returns
    --------------
    digraph
        Graphviz DiGraph object
    """
    if parameters is None:
        parameters = {}

    format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, constants.DEFAULT_FORMAT_GVIZ_VIEW
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, "LR")

    include_node_attributes = exec_utils.get_param_value(
        Parameters.INCLUDE_NODE_ATTRIBUTES, parameters, True
    )
    include_edge_attributes = exec_utils.get_param_value(
        Parameters.INCLUDE_EDGE_ATTRIBUTES, parameters, True
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    viz = graphviz.Digraph(
        "networkx_digraph",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor},
    )

    nodes_dict = {}
    for node in G.nodes:
        node_id = str(uuid.uuid4())
        nodes_dict[node] = node_id
        label = str(node)

        if include_node_attributes and G.nodes[node]:
            label += "\n" + str(G.nodes[node]).replace(", ", ",\n")

        viz.node(node_id, label=label, shape="box")

    for edge in G.edges:
        label = " "
        if include_edge_attributes and G.edges[edge]:
            label = str(G.edges[edge]).replace(", ", ",\n")

        viz.edge(nodes_dict[edge[0]], nodes_dict[edge[1]], label=label)

    viz.attr(rankdir=rankdir)
    viz.format = format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.util import exec_utils
from enum import Enum
from pm4py.algo.conformance.declare.variants import classic
from pm4py.objects.log.obj import EventLog
import pandas as pd
from typing import Union, Dict, Optional, Any, List


class Variants(Enum):
    CLASSIC = classic


def apply(
    log: Union[EventLog, pd.DataFrame],
    model: Dict[str, Dict[Any, Dict[str, int]]],
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> List[Dict[str, Any]]:
    """
    Applies conformance checking against a DECLARE model.

    Parameters
    --------------
    log
        Event log / Pandas dataframe
    model
        DECLARE model
    variant
        Variant to be used:
        - Variants.CLASSIC
    parameters
        Variant-specific parameters

    Returns
    -------------
    lst_conf_res
        List containing for every case a dictionary with different keys:
        - no_constr_total => the total number of constraints of the DECLARE model
        - deviations => a list of deviations
        - no_dev_total => the total number of deviations
        - dev_fitness => the fitness (1 - no_dev_total / no_constr_total)
        - is_fit => True if the case is perfectly fit
    """
    return exec_utils.get_variant(variant).apply(log, model, parameters)


def get_diagnostics_dataframe(
    log, conf_result, variant=Variants.CLASSIC, parameters=None
) -> pd.DataFrame:
    """
    Gets the diagnostics dataframe from a log and the results
    of DECLARE-based conformance checking

    Parameters
    --------------
    log
        Event log
    conf_result
        Results of conformance checking
    variant
        Variant to be used:
        - Variants.CLASSIC
    parameters
        Variant-specific parameters

    Returns
    --------------
    diagn_dataframe
        Diagnostics dataframe
    """
    return exec_utils.get_variant(variant).get_diagnostics_dataframe(
        log, conf_result, parameters
    )
